"""
Composio SDK tools.
"""

import base64
import binascii
import hashlib
import importlib
import inspect
import itertools
import json
import os
import time
import typing as t
import warnings
from datetime import datetime
from functools import wraps
from importlib.util import find_spec
from pathlib import Path

import typing_extensions as te
from pydantic import BaseModel

from composio import Action, ActionType, App, AppType, TagType
from composio.client import Composio, Entity
from composio.client.collections import (
    AUTH_SCHEMES,
    ActionModel,
    AppAuthScheme,
    AppModel,
    AuthSchemeField,
    AuthSchemeType,
    ConnectedAccountModel,
    ConnectionParams,
    ConnectionRequestModel,
    CustomAuthObject,
    CustomAuthParameter,
    ExpectedFieldInput,
    FileType,
    IntegrationModel,
    SuccessExecuteActionResponseModel,
    TriggerModel,
    TriggerSubscription,
)
from composio.client.enums import TriggerType
from composio.client.enums.base import EnumStringNotFound
from composio.client.exceptions import ComposioClientError, HTTPError, NoItemsFound
from composio.constants import (
    DEFAULT_ENTITY_ID,
    ENV_COMPOSIO_API_KEY,
    LOCAL_CACHE_DIRECTORY,
    LOCAL_OUTPUT_FILE_DIRECTORY_NAME,
    USER_DATA_FILE_NAME,
)
from composio.exceptions import ApiKeyNotProvidedError, ComposioSDKError
from composio.storage.user import UserData
from composio.tools.base.abs import tool_registry
from composio.tools.base.local import LocalAction
from composio.tools.env.base import (
    ENV_GITHUB_ACCESS_TOKEN,
    Workspace,
    WorkspaceConfigType,
)
from composio.tools.env.factory import HostWorkspaceConfig, WorkspaceFactory
from composio.tools.local import load_local_tools
from composio.tools.local.handler import LocalClient
from composio.utils.enums import get_enum_key
from composio.utils.logging import LogIngester, LogLevel, WithLogger
from composio.utils.url import get_api_url_base


T = te.TypeVar("T")
P = te.ParamSpec("P")

_KeyType = t.Union[AppType, ActionType]
_CallableType = t.Callable[[t.Dict], t.Dict]

MetadataType = t.Dict[_KeyType, t.Dict]
ParamType = t.TypeVar("ParamType")
ProcessorType = te.Literal["pre", "post", "schema"]


class IntegrationParams(te.TypedDict):

    integration_id: str
    auth_scheme: str
    expected_params: t.List[ExpectedFieldInput]


class ProcessorsType(te.TypedDict):
    """Request and response processors."""

    pre: te.NotRequired[t.Dict[_KeyType, _CallableType]]
    """Request processors."""

    post: te.NotRequired[t.Dict[_KeyType, _CallableType]]
    """Response processors."""

    schema: te.NotRequired[t.Dict[_KeyType, _CallableType]]
    """Schema processors"""


def _check_agentops() -> bool:
    """Check if AgentOps is installed and initialized."""
    if find_spec("agentops") is None:
        return False
    import agentops  # pylint: disable=import-outside-toplevel # type: ignore

    return agentops.get_api_key() is not None


def _record_action_if_available(func: t.Callable[P, T]) -> t.Callable[P, T]:
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        if _check_agentops():
            import agentops  # pylint: disable=import-outside-toplevel # type: ignore

            action_name = str(kwargs.get("action", "unknown_action"))
            return agentops.record_action(action_name)(func)(self, *args, **kwargs)
        return func(self, *args, **kwargs)  # type: ignore

    return wrapper  # type: ignore


def load_action(
    client: Composio, value, warn=True
) -> Action:  # pylint: disable=used-prior-global-declaration
    global Action
    try:
        return Action(value=value, warn=warn)
    except EnumStringNotFound as e:
        # run update apps, and reload actions
        from composio.cli.apps import (  # pylint: disable=import-outside-toplevel
            update_actions,
            update_apps,
        )

        apps = update_apps(client)
        update_actions(client, apps)
        action_enum_module = inspect.getmodule(Action)
        if action_enum_module is None:
            raise RuntimeError("Error reloading `Action` enum class") from e
        reloaded_action_module = importlib.reload(action_enum_module)
        Action = reloaded_action_module.Action  # type: ignore

    return Action(value=value, warn=warn)


class ComposioToolSet(WithLogger):  # pylint: disable=too-many-public-methods
    """Composio toolset."""

    _custom_auth: t.Dict[App, CustomAuthObject]

    _connected_accounts: t.Optional[t.List[ConnectedAccountModel]] = None
    _remote_client: t.Optional[Composio] = None
    _workspace: t.Optional[Workspace] = None

    _runtime: str = "composio"
    _description_char_limit: int = 1024
    _log_ingester_client: t.Optional[LogIngester] = None

    def __init_subclass__(
        cls,
        runtime: t.Optional[str] = None,
        description_char_limit: t.Optional[int] = None,
    ) -> None:
        if runtime is None:
            warnings.warn(
                f"runtime is not set on {cls.__name__}, using 'composio' as default"
            )
        cls._runtime = runtime or "composio"

        if description_char_limit is None:
            warnings.warn(
                f"description_char_limit is not set on {cls.__name__}, using 1024 as default"
            )
        cls._description_char_limit = description_char_limit or 1024

    def __init__(
        self,
        api_key: t.Optional[str] = None,
        base_url: t.Optional[str] = None,
        entity_id: str = DEFAULT_ENTITY_ID,
        workspace_id: t.Optional[str] = None,
        workspace_config: t.Optional[WorkspaceConfigType] = None,
        metadata: t.Optional[MetadataType] = None,
        processors: t.Optional[ProcessorsType] = None,
        output_in_file: bool = False,
        logging_level: LogLevel = LogLevel.INFO,
        output_dir: t.Optional[Path] = None,
        verbosity_level: t.Optional[int] = None,
        connected_account_ids: t.Optional[t.Dict[AppType, str]] = None,
        **kwargs: t.Any,
    ) -> None:
        """
        Initialize composio toolset

        :param api_key: Composio API key
        :param base_url: Base URL for the Composio API server
        :param runtime: Name of the framework runtime, eg. openai, crewai...
        :param output_in_file: Whether to output the result to a file.
        :param entity_id: The ID of the entity to execute the action on.
            Defaults to "default".
        :param workspace_env: Environment where actions should be executed,
            you can choose from `host`, `docker`, `flyio` and `e2b`.
        :param workspace_id: Workspace ID for loading an existing workspace
        :param metadata: Additional metadata for executing an action or an
            action which belongs to a specific app. The additional metadata
            needs to be JSON serialisable dictionary. For example

            ```python
            toolset = ComposioToolSet(
                ...,
                metadata={
                    App.IMAGEANALYSER: {
                        "base_url": "https://image.analyser/api",
                    },
                    Action.IMAGEANALYSER_GTP4:{
                        "openai_api_key": "sk-proj-somekey",
                    }
                }
            )
            ```
        :param processors: Request and response processors, use these to
            pre-process requests before executing an action and post-process
            the response after an action has been executed. The processors can
            be defined at app and action level. The order of execution will be

            `App pre-processor -> Action pre-processor -> execute action -> Action post-processor -> App post-processor`

            Heres and example of a request pre-processor

            ```python
            def _add_cwd_if_missing(request: t.Dict) -> t.Dict:
                if "cwd" not in request:
                    request["cwd"] = "~/project"
                return request

            def _sanitise_file_search_request(request: t.Dict) -> t.Dict:
                if ".tox" not in request["exclude"]:
                    request["exclude"].append(".tox")
                return request

            def _limit_file_search_response(response: t.Dict) -> t.Dict:
                if len(response["results"]) > 100:
                    response["results"] = response["results"][:100]
                return response

            toolset = ComposioToolSet(
                ...,
                processors={
                    "pre": {
                        App.FILETOOL: _add_cwd_if_missing,
                        Action.FILETOOL_SEARCH: _sanitise_file_search_request,
                    },
                    "post": {
                        Action.FILETOOL_SEARCH: _limit_file_search_response,
                    }
                }
            )
            ```
        :param verbosity_level: This defines the size of the log object that will
            be printed on the console.
        :param connection_ids: Use this to define connection IDs to use when executing
            an action for a specific app.
        """
        super().__init__(
            logging_level=logging_level,
            verbosity_level=verbosity_level,
        )
        self.logger.info(
            f"Logging is set to {self._logging_level}, "
            "use `logging_level` argument or "
            "`COMPOSIO_LOGGING_LEVEL` change this"
        )
        self.entity_id = entity_id
        self.output_in_file = output_in_file
        self.output_dir = (
            output_dir or LOCAL_CACHE_DIRECTORY / LOCAL_OUTPUT_FILE_DIRECTORY_NAME
        )
        self._ensure_output_dir_exists()

        self._base_url = base_url or get_api_url_base()
        try:
            self._api_key = (
                api_key
                or os.environ.get(ENV_COMPOSIO_API_KEY)
                or UserData.load(LOCAL_CACHE_DIRECTORY / USER_DATA_FILE_NAME).api_key
            )
        except FileNotFoundError:
            self._api_key = None
            self.logger.debug("`api_key` is not set when initializing toolset.")

        if processors is not None:
            warnings.warn(
                "Setting 'processors' on the ToolSet is deprecated, they should"
                "be provided to the 'get_tools()' method instead.",
                DeprecationWarning,
                stacklevel=2,
            )
            self._processors: ProcessorsType = processors
        else:
            self._processors = {"post": {}, "pre": {}, "schema": {}}

        self._metadata = metadata or {}
        self._workspace_id = workspace_id
        self._workspace_config = workspace_config
        self._local_client = LocalClient()
        self._custom_auth = {}

        if len(kwargs) > 0:
            self.logger.info(f"Extra kwargs while initializing toolset: {kwargs}")

        self.logger.debug("Loading local tools")
        load_local_tools()

        self._connected_account_ids = self._validating_connection_ids(
            connected_account_ids=connected_account_ids or {}
        )

    def _validating_connection_ids(
        self,
        connected_account_ids: t.Dict[AppType, str],
    ) -> t.Dict[App, str]:
        """Validate connection IDs."""
        valid = {}
        invalid = []
        entity = self.client.get_entity(id=self.entity_id)
        for app, connected_account_id in connected_account_ids.items():
            self.logger.debug(f"Validating {app} {connected_account_id=}")
            try:
                entity.get_connection(
                    app=app,
                    connected_account_id=connected_account_id,
                )
                valid[App(app)] = connected_account_id
            except HTTPError:
                invalid.append((str(app), connected_account_id))

        if len(invalid) == 0:
            return valid

        raise ComposioSDKError(message=f"Invalid connected accounts found: {invalid}")

    def _get_connected_account(self, action: ActionType) -> t.Optional[str]:
        return self._connected_account_ids.get(App(Action(action).app))

    def _try_get_github_access_token_for_current_entity(self) -> t.Optional[str]:
        """Try and get github access token for current entiry."""
        from_env = os.environ.get(f"_COMPOSIO_{ENV_GITHUB_ACCESS_TOKEN}")
        if from_env is not None:
            self.logger.debug("Using composio github access token")
            return from_env

        self.logger.debug(f"Trying to get github access token for {self.entity_id=}")
        try:
            account = self.client.get_entity(id=self.entity_id).get_connection(
                app=App.GITHUB
            )
            token = (
                self.client.connected_accounts.get(connection_id=account.id)
                .connectionParams.headers["Authorization"]  # type: ignore
                .replace("Bearer ", "")
            )
            self.logger.debug(
                f"Using `{token}` with scopes: {account.connectionParams.scope}"
            )
            return token
        except ComposioClientError:
            return None

    @property
    def _log_ingester(self) -> LogIngester:
        if self._log_ingester_client is None:
            self._log_ingester_client = LogIngester()
        return self._log_ingester_client

    @property
    def api_key(self) -> str:
        if self._api_key is None:
            raise ApiKeyNotProvidedError()
        return self._api_key

    @property
    def client(self) -> Composio:
        if self._remote_client is None:
            self._remote_client = Composio(
                api_key=self._api_key,
                base_url=self._base_url,
                runtime=self._runtime,
            )
        self._remote_client.local = self._local_client
        return self._remote_client

    @property
    def workspace(self) -> Workspace:
        """Workspace for this toolset instance."""
        if self._workspace is not None:
            return self._workspace

        if self._workspace_id is not None:
            self._workspace = WorkspaceFactory.get(id=self._workspace_id)
            return self._workspace

        workspace_config = self._workspace_config or HostWorkspaceConfig()
        if workspace_config.composio_api_key is None:
            workspace_config.composio_api_key = self.api_key

        if workspace_config.composio_base_url is None:
            workspace_config.composio_base_url = self._base_url

        if workspace_config.github_access_token is None:
            workspace_config.github_access_token = (
                self._try_get_github_access_token_for_current_entity()
            )

        self._workspace = WorkspaceFactory.new(config=workspace_config)
        return self._workspace

    def set_workspace_id(self, workspace_id: str) -> None:
        self._workspace_id = workspace_id
        if self._workspace is not None:
            self._workspace = WorkspaceFactory.get(id=workspace_id)

    def add_auth(
        self,
        app: AppType,
        parameters: t.List[CustomAuthParameter],
        base_url: t.Optional[str] = None,
        body: t.Optional[t.Dict] = None,
    ) -> None:
        self._custom_auth[App(app)] = CustomAuthObject(
            body=body or {},
            base_url=base_url,
            parameters=parameters,
        )

    def check_connected_account(self, action: ActionType) -> None:
        """Check if connected account is required and if required it exists or not."""
        action = load_action(self.client, action)
        if action.no_auth or action.is_runtime:
            return

        if App(action.app) in self._custom_auth:
            return

        if self._connected_accounts is None:
            self._connected_accounts = t.cast(
                t.List[ConnectedAccountModel],
                self.client.connected_accounts.get(),
            )

        if action.app not in [
            connection.appUniqueId for connection in self._connected_accounts
        ]:
            raise ComposioSDKError(
                f"No connected account found for app `{action.app}`; "
                f"Run `composio add {action.app.lower()}` to fix this"
            )

    def _execute_local(
        self,
        action: Action,
        params: t.Dict,
        metadata: t.Optional[t.Dict] = None,
        entity_id: t.Optional[str] = None,
    ) -> t.Dict:
        """Execute a local action."""
        metadata = metadata or {}
        metadata["_toolset"] = self
        response = self.workspace.execute_action(
            action=action,
            request_data=params,
            metadata={
                **(metadata or {}),
                "entity_id": entity_id or self.entity_id,
            },
        )

        if isinstance(response, BaseModel):
            response = response.model_dump()

        self._log_ingester.log(
            connection_id=None,
            provider_name=action.app,
            action_name=action.name,
            request=params,
            response=response,
            is_error=not response.get("successful", False),
            session_id=self.workspace.id,
        )

        return response

    def _execute_remote(
        self,
        action: Action,
        params: t.Dict,
        entity_id: str = DEFAULT_ENTITY_ID,
        connected_account_id: t.Optional[str] = None,
        session_id: t.Optional[str] = None,
        text: t.Optional[str] = None,
    ) -> t.Dict:
        """Execute a remote action."""
        auth = self._custom_auth.get(App(action.app))
        if auth is None:
            self.check_connected_account(action=action)

        output = self.client.get_entity(id=entity_id).execute(
            action=action,
            params=params,
            connected_account_id=connected_account_id,
            session_id=session_id,
            text=text,
            auth=auth,
        )

        if self.output_in_file:
            return self._write_to_file(
                action=action,
                output=output,
                entity_id=entity_id,
            )

        try:
            self.logger.debug("Trying to validate success response model")
            success_response_model = SuccessExecuteActionResponseModel.model_validate(
                output
            )
        except Exception:
            self.logger.debug("Failed to validate success response model")
            return output

        return self._save_var_files(
            file_name_prefix=f"{action.name}_{entity_id}_{time.time()}",
            success_response_model=success_response_model,
        )

    def _ensure_output_dir_exists(self):
        """Ensure the output directory exists."""
        if not self.output_dir.exists():
            self.output_dir.mkdir(parents=True, exist_ok=True)

    def _save_var_files(
        self,
        file_name_prefix: str,
        success_response_model: SuccessExecuteActionResponseModel,
    ) -> dict:
        error = success_response_model.error
        resp_data = success_response_model.data
        is_invalid_file = False
        for key, val in resp_data.items():
            try:
                file_model = FileType.model_validate(val)
                self._ensure_output_dir_exists()

                local_filepath = (
                    self.output_dir
                    / f"{file_name_prefix}_{file_model.name.replace('/', '_')}"
                )

                _write_file(
                    local_filepath, base64.urlsafe_b64decode(file_model.content)
                )

                resp_data[key] = str(local_filepath)
            except binascii.Error:
                is_invalid_file = True
                resp_data[key] = "Invalid File! Unable to decode."
            except Exception:
                pass

        if is_invalid_file is True and error is None:
            success_response_model.error = "Execution failed"
        success_response_model.data = resp_data
        return success_response_model.model_dump()

    def _write_to_file(
        self,
        action: Action,
        output: t.Dict,
        entity_id: str = DEFAULT_ENTITY_ID,
    ) -> t.Dict:
        """Write output to a file."""
        filename = hashlib.sha256(
            f"{action.name}-{entity_id}-{time.time()}".encode()
        ).hexdigest()
        self._ensure_output_dir_exists()
        outfile = self.output_dir / filename
        self.logger.info(f"Writing output to: {outfile}")
        _write_file(outfile, json.dumps(output))
        return {
            "message": f"output written to {outfile.resolve()}",
            "file": str(outfile.resolve()),
        }

    def _serialize_execute_params(self, param: ParamType) -> ParamType:
        """Returns a serialized version of the parameters object."""
        if param is None:
            return param  # type: ignore

        if isinstance(param, (int, float, str, bool)):
            return param  # type: ignore

        if isinstance(param, BaseModel):
            return param.model_dump_json(exclude_none=True)  # type: ignore

        if isinstance(param, list):
            return [self._serialize_execute_params(p) for p in param]  # type: ignore

        if isinstance(param, dict):
            return {
                key: self._serialize_execute_params(val)  # type: ignore
                for key, val in param.items()
            }

        raise ValueError(
            "Invalid value found for execute parameters"
            f"\ntype={type(param)} \nvalue={param}"
        )

    def _get_metadata(self, key: _KeyType) -> t.Dict:
        metadata = self._metadata.get(key)  # type: ignore
        if metadata is not None:
            return metadata

        try:
            return self._metadata.get(Action(t.cast(ActionType, key)), {})  # type: ignore
        except EnumStringNotFound:
            return self._metadata.get(App(t.cast(AppType, key)), {})  # type: ignore

    def _add_metadata(self, action: Action, metadata: t.Optional[t.Dict]) -> t.Dict:
        metadata = metadata or {}
        metadata.update(self._get_metadata(key=App(action.app)))
        metadata.update(self._get_metadata(key=action))
        return metadata

    def _get_processor(
        self,
        key: _KeyType,
        type_: te.Literal["post", "pre", "schema"],
    ) -> t.Optional[_CallableType]:
        """Get processor for given app or action"""
        processor = self._processors.get(type_, {}).get(key)  # type: ignore
        if processor is not None:
            return processor

        try:
            return self._processors.get(type_, {}).get(Action(t.cast(ActionType, key)))  # type: ignore
        except EnumStringNotFound:
            return self._processors.get(type_, {}).get(App(t.cast(AppType, key)))  # type: ignore

    def _process(
        self,
        key: _KeyType,
        data: t.Dict,
        type_: te.Literal["pre", "post", "schema"],
    ) -> t.Dict:
        processor = self._get_processor(key=key, type_=type_)
        if processor is not None:
            self.logger.info(
                f"Running {'request' if type_ == 'pre' else 'response' if type_ == 'post' else 'schema'}"
                f" through: {processor.__name__}"
            )
            data = processor(data)
            # Users may not respect our type annotations and return something that isn't a dict.
            # If that happens we should show a friendly error message.
            if not isinstance(data, t.Dict):
                warnings.warn(
                    f"Expected {type_}-processor to return 'dict', got {type(data).__name__!r}",
                    stacklevel=2,
                )
        return data

    def _process_request(self, action: Action, request: t.Dict) -> t.Dict:
        return self._process(
            key=action,
            data=self._process(
                key=App(action.app),
                data=request,
                type_="pre",
            ),
            type_="pre",
        )

    def _process_respone(self, action: Action, response: t.Dict) -> t.Dict:
        return self._process(
            key=App(action.app),
            data=self._process(
                key=action,
                data=response,
                type_="post",
            ),
            type_="post",
        )

    def _process_schema_properties(self, action: Action, properties: t.Dict) -> t.Dict:
        return self._process(
            key=App(action.app),
            data=self._process(
                key=action,
                data=properties,
                type_="schema",
            ),
            type_="schema",
        )

    def _merge_processors(self, processors: ProcessorsType) -> None:
        for processor_type in self._processors.keys():
            if processor_type not in processors:
                continue

            processor_type = t.cast(ProcessorType, processor_type)
            new_processors = processors[processor_type]

            if processor_type in self._processors:
                existing_processors = self._processors[processor_type]
            else:
                existing_processors = {}
                self._processors[processor_type] = existing_processors

            existing_processors.update(new_processors)

    @_record_action_if_available
    def execute_action(
        self,
        action: ActionType,
        params: dict,
        metadata: t.Optional[t.Dict] = None,
        entity_id: t.Optional[str] = None,
        connected_account_id: t.Optional[str] = None,
        text: t.Optional[str] = None,
        *,
        processors: t.Optional[ProcessorsType] = None,
    ) -> t.Dict:
        """
        Execute an action on a given entity.

        :param action: Action to execute
        :param params: The parameters to pass to the action
        :param entity_id: The ID of the entity to execute the action on. Defaults to "default"
        :param text: Extra text to use for generating function calling metadata
        :param metadata: Metadata for executing local action
        :param connected_account_id: Connection ID for executing the remote action
        :return: Output object from the function call
        """
        action = Action(action)
        params = self._serialize_execute_params(param=params)
        if processors is not None:
            self._merge_processors(processors)

        if not action.is_runtime:
            params = self._process_request(action=action, request=params)
            metadata = self._add_metadata(action=action, metadata=metadata)
            connected_account_id = connected_account_id or self._get_connected_account(
                action=action
            )

        self.logger.info(
            f"Executing `{action.slug}` with {params=} and {metadata=} {connected_account_id=}"
        )
        response = (
            self._execute_local(
                action=action,
                params=params,
                metadata=metadata,
                entity_id=entity_id,
            )
            if action.is_local
            else self._execute_remote(
                action=action,
                params=params,
                entity_id=entity_id or self.entity_id,
                connected_account_id=connected_account_id,
                text=text,
                session_id=self.workspace.id,
            )
        )
        response = (
            response
            if action.is_runtime
            else self._process_respone(action=action, response=response)
        )
        self.logger.info(f"Got {response=} from {action=} with {params=}")
        return response

    @t.overload
    def execute_request(
        self,
        endpoint: str,
        method: str,
        *,
        body: t.Optional[t.Dict] = None,
        parameters: t.Optional[t.List[CustomAuthParameter]] = None,
        connection_id: t.Optional[str] = None,
    ) -> t.Dict:
        pass

    @t.overload
    def execute_request(
        self,
        endpoint: str,
        method: str,
        *,
        body: t.Optional[t.Dict] = None,
        parameters: t.Optional[t.List[CustomAuthParameter]] = None,
        app: t.Optional[AppType] = None,
    ) -> t.Dict:
        pass

    def execute_request(
        self,
        endpoint: str,
        method: str,
        *,
        body: t.Optional[t.Dict] = None,
        parameters: t.Optional[t.List[CustomAuthParameter]] = None,
        connection_id: t.Optional[str] = None,
        app: t.Optional[AppType] = None,
    ) -> t.Dict:
        """
        Execute a proxy request to a connected account.

        :param endpoint: API endpoint to call
        :param method: HTTP method to use (GET, POST, etc.)
        :param body: Request body data
        :param parameters: Additional auth parameters
        :param connection_id: ID of the connected account
        :param app: App type to use for connection lookup

        :returns: Response from the proxy request
        :raises: ComposioSDKError: If neither connection_id nor app is provided
        """
        if app is not None and connection_id is None:
            connection_id = (
                self.get_entity(id=self.entity_id).get_connection(app=app).id
            )

        if connection_id is None:
            raise ComposioSDKError(
                "Please provide connection id or app name to execute a request"
            )

        self.logger.info(
            f"Executing request to {endpoint} with method={method}, connection_id={connection_id}"
        )
        response = self.client.actions.request(
            connection_id=connection_id,
            body=body,
            method=method,
            endpoint=endpoint,
            parameters=parameters,
        )
        self.logger.info(f"Got {response=}")
        return response

    def validate_tools(
        self,
        apps: t.Optional[t.Sequence[AppType]] = None,
        actions: t.Optional[t.Sequence[ActionType]] = None,
        tags: t.Optional[t.Sequence[TagType]] = None,
    ) -> None:
        # NOTE: This an experimental, can convert to decorator for more convinience
        if not apps and not actions and not tags:
            return
        self.workspace.check_for_missing_dependencies(
            apps=apps,
            actions=actions,
            tags=tags,
        )

    def get_action_schemas(
        self,
        apps: t.Optional[t.Sequence[AppType]] = None,
        actions: t.Optional[t.Sequence[ActionType]] = None,
        tags: t.Optional[t.Sequence[TagType]] = None,
        *,
        check_connected_accounts: bool = True,
    ) -> t.List[ActionModel]:
        runtime_actions = t.cast(
            t.List[t.Type[LocalAction]],
            [action for action in actions or [] if hasattr(action, "run_on_shell")],
        )
        actions = t.cast(
            t.List[Action],
            [
                Action(action)
                for action in actions or []
                if action not in runtime_actions
            ],
        )
        apps = t.cast(t.List[App], [App(app) for app in apps or []])
        items: t.List[ActionModel] = []

        local_actions = [action for action in actions if action.is_local]
        local_apps = [app for app in apps if app.is_local]
        if len(local_actions) > 0 or len(local_apps) > 0:
            items += [
                ActionModel(**item)
                for item in self._local_client.get_action_schemas(
                    apps=local_apps,
                    actions=local_actions,
                    tags=tags,
                )
            ]

        remote_actions = [action for action in actions if not action.is_local]
        remote_apps = [app for app in apps if not app.is_local]
        if len(remote_actions) > 0 or len(remote_apps) > 0:
            remote_items = self.client.actions.get(
                apps=remote_apps,
                actions=remote_actions,
                tags=tags,
            )
            if check_connected_accounts:
                for item in remote_items:
                    self.check_connected_account(action=item.name)
            else:
                warnings.warn(
                    "Not verifying connected accounts for apps."
                    " Actions may fail when the Agent tries to use them.",
                    UserWarning,
                )
            items = items + remote_items

        for act in runtime_actions:
            schema = act.schema()
            schema["name"] = act.enum
            items.append(ActionModel(**schema).model_copy(deep=True))

        for item in items:
            item = self._process_schema(item)

            # This is to support anthropic-claude
            if item.name == Action.ANTHROPIC_BASH_COMMAND.slug:
                item.name = "bash"

            if item.name == Action.ANTHROPIC_COMPUTER.slug:
                item.name = "computer"

            if item.name == Action.ANTHROPIC_TEXT_EDITOR.slug:
                item.name = "str_replace_editor"

        return items

    def _process_schema(self, action_item: ActionModel) -> ActionModel:
        required_params = action_item.parameters.required or []
        for param_name, param_details in action_item.parameters.properties.items():
            if param_details.get("title") == "FileType" and all(
                fprop in param_details.get("properties", {})
                for fprop in ("name", "content")
            ):
                action_item.parameters.properties[param_name].pop("properties")
                action_item.parameters.properties[param_name] = {
                    "default": param_details.get("default"),
                    "type": "string",
                    "format": "file-path",
                    "description": f"File path to {param_details.get('description', '')}",
                }
            elif param_details.get("allOf", [{}])[0].get("title") == "FileType" and all(
                fprop in param_details.get("allOf", [{}])[0].get("properties", {})
                for fprop in ("name", "content")
            ):
                action_item.parameters.properties[param_name].pop("allOf")
                action_item.parameters.properties[param_name].update(
                    {
                        "default": param_details.get("default"),
                        "type": "string",
                        "format": "file-path",
                        "description": f"File path to {param_details.get('description', '')}",
                    }
                )
            elif param_details.get("type") in [
                "string",
                "integer",
                "number",
                "boolean",
            ]:
                ext = f'Please provide a value of type {param_details["type"]}.'
                description = param_details.get("description", "").rstrip(".")
                param_details["description"] = (
                    f"{description}. {ext}" if description else ext
                )

            if param_name in required_params:
                description = param_details.get("description", "")
                if description:
                    param_details["description"] = (
                        f"{description.rstrip('.')}. This parameter is required."
                    )
                else:
                    param_details["description"] = "This parameter is required."

        if action_item.description is not None:
            action_item.description = action_item.description[
                : self._description_char_limit
            ]
        action_item.parameters.properties = self._process_schema_properties(
            action=Action(action_item.name.upper()),
            properties=action_item.parameters.properties,
        )
        return action_item

    def create_trigger_listener(self, timeout: float = 15.0) -> TriggerSubscription:
        """Create trigger subscription."""
        return self.client.triggers.subscribe(timeout=timeout)

    def find_actions_by_use_case(
        self,
        *apps: AppType,
        use_case: str,
    ) -> t.List[Action]:
        """
        Find actions by specified use case.

        :param apps: List of apps to search.
        :param use_case: String describing the use case.
        :return: A list of actions matching the relevant use case.
        """
        actions = self.client.actions.get(
            apps=[App(app) for app in apps],
            use_case=use_case,
            allow_all=True,
        )
        return [
            Action(value=get_enum_key(name=action.name).lower()) for action in actions
        ]

    def find_actions_by_tags(
        self,
        *apps: AppType,
        tags: t.List[str],
    ) -> t.List[Action]:
        """
        Find actions by specified use case.

        :param apps: List of apps to search.
        :param use_case: String describing the use case.
        :return: A list of actions matching the relevant use case.
        """
        if len(tags) == 0:
            raise ComposioClientError(
                "Please provide at least one tag to perform search"
            )

        if len(apps) > 0:
            return list(
                itertools.chain(
                    *[list(App(app).get_actions(tags=tags)) for app in apps]
                )
            )

        actions = []
        for action in Action.all():
            if any(tag in action.tags for tag in tags):
                actions.append(action)
        return actions

    def get_agent_instructions(
        self,
        apps: t.Optional[t.Sequence[AppType]] = None,
        actions: t.Optional[t.Sequence[ActionType]] = None,
        tags: t.Optional[t.Sequence[TagType]] = None,
    ) -> str:
        """
        Generate a formatted string with instructions for agents based on the provided apps, actions, and tags.

        This function compiles a list of available tools from the specified apps, actions, and tags,
        and formats them into a human-readable string that can be used as instructions for agents.

        :param apps: Optional sequence of AppType to include in the search.
        :param actions: Optional sequence of ActionType to include in the search.
        :param tags: Optional sequence of TagType to filter the actions.
        :return: A formatted string with instructions for agents.
        """
        # Retrieve schema information for the given apps, actions, and tags
        schema_list = [
            schema.model_dump()
            for schema in (
                self.get_action_schemas(apps=apps, tags=tags)
                + self.get_action_schemas(actions=actions)
            )
        ]
        schema_info = [
            (schema_obj["appName"], schema_obj["name"]) for schema_obj in schema_list
        ]

        # Helper function to format a list of items into a string
        def format_list(items):
            if not items:
                return ""
            if len(items) == 1:
                return items[0]
            return ", ".join(items[:-2] + [" and ".join(items[-2:])])

        # Organize the schema information by app name
        action_dict: t.Dict[str, t.List] = {}
        for appName, name in schema_info:
            if appName not in action_dict:
                action_dict[appName] = []
            action_dict[appName].append(name)

        # Format the schema information into a human-readable string
        formatted_schema_info = (
            "You have various tools, among which "
            + ", ".join(
                [
                    f"for interacting with **{appName}** you might use {format_list(action_items)} tools"
                    for appName, action_items in action_dict.items()
                ]
            )
            + ". Whichever tool is useful to execute your task, use that with proper parameters."
        )
        return formatted_schema_info

    def get_auth_params(
        self,
        app: t.Optional[AppType] = None,
        connection_id: t.Optional[str] = None,
        entity_id: t.Optional[str] = None,
    ) -> t.Optional[ConnectionParams]:
        """Get authentication parameters for given app."""
        if app is None and connection_id is None:
            raise ComposioSDKError("Both `app` and `connection_id` cannot be `None`")

        try:
            connection_id = (
                connection_id
                or self.client.get_entity(id=entity_id or self.entity_id)
                .get_connection(app=app)
                .id
            )
            return self.client.connected_accounts.info(connection_id=connection_id)
        except ComposioClientError:
            return None

    def get_auth_schemes(self, app: AppType) -> t.List[AppAuthScheme]:
        """Get the list of auth schemes for an app."""
        return self.client.apps.get(name=str(app)).auth_schemes or []

    def get_app(self, app: AppType) -> AppModel:
        return self.client.apps.get(name=str(App(app)))

    def get_apps(
        self,
        no_auth: t.Optional[bool] = None,
        include_local: bool = True,
    ) -> t.List[AppModel]:
        apps = self.client.apps.get()
        if no_auth is not None:
            apps = [a for a in apps if a.no_auth is no_auth]

        if include_local:
            for app in tool_registry["local"].values():
                apps.append(
                    AppModel(
                        name=app.name,
                        key=app.name,
                        appId=app.name,
                        description=app.description,
                        categories=["local"],
                        meta={},
                        no_auth=True,
                    )
                )
        return apps

    def get_action(self, action: ActionType) -> ActionModel:
        return self.client.actions.get(actions=[action]).pop()

    def get_trigger(self, trigger: TriggerType) -> TriggerModel:
        return self.client.triggers.get(trigger_names=[trigger]).pop()

    def get_integration(self, id: str) -> IntegrationModel:
        return self.client.integrations.get(id=id)

    def get_integrations(
        self,
        app: t.Optional[AppType] = None,
        auth_scheme: t.Optional[AuthSchemeType] = None,
    ) -> t.List[IntegrationModel]:
        integrations = self.client.integrations.get()
        if app is not None:
            app = str(app).lower()
            integrations = [i for i in integrations if i.appName.lower() == app]

        if auth_scheme is not None:
            integrations = [i for i in integrations if i.authScheme == auth_scheme]

        return integrations

    def get_connected_account(self, id: str) -> ConnectedAccountModel:
        return self.client.connected_accounts.get(connection_id=id)

    def get_connected_accounts(self) -> t.List[ConnectedAccountModel]:
        return self.client.connected_accounts.get()

    def get_entity(self, id: t.Optional[str] = None) -> Entity:
        """Get entity object for given ID."""
        return self.client.get_entity(id=id or self.entity_id)

    def get_auth_scheme_for_app(
        self,
        app: t.Optional[AppType] = None,
        auth_scheme: t.Optional[AuthSchemeType] = None,
    ) -> AppAuthScheme:
        auth_schemes = {
            scheme.auth_mode: scheme
            for scheme in self.client.apps.get(name=str(app)).auth_schemes or []
        }

        if auth_scheme is not None and auth_scheme not in auth_schemes:
            raise ComposioSDKError(
                message=f"Auth scheme `{auth_scheme}` not found for app `{app}`"
            )

        if auth_scheme is not None:
            return auth_schemes[auth_scheme]

        for scheme in AUTH_SCHEMES:
            if scheme in auth_schemes:
                scheme = t.cast(AuthSchemeType, scheme)
                return auth_schemes[scheme]

        raise ComposioSDKError(
            message=(
                f"Error getting expected params for {app=}, {auth_scheme=}, "
                f"available_schems={list(auth_schemes)}"
            )
        )

    def _get_expected_params_from_integration_id(self, id: str) -> IntegrationParams:
        integration = self.get_integration(id=id)
        return {
            "integration_id": integration.id,
            "auth_scheme": integration.authScheme,
            "expected_params": integration.expectedInputFields,
        }

    def _get_integration_for_app(
        self,
        app: AppType,
        auth_scheme: t.Optional[str] = None,
    ) -> IntegrationModel:
        for integration in sorted(self.get_integrations(), key=lambda x: x.createdAt):
            if integration.appName.lower() == str(app).lower():
                if (
                    auth_scheme is not None
                    and integration.authScheme.lower() != auth_scheme.lower()
                ):
                    continue
                return self.get_integration(id=integration.id)
        raise NoItemsFound(f"No integration found for `{app}`")

    def _get_expected_params_from_app(
        self,
        app: AppType,
        auth_scheme: t.Optional[str] = None,
    ) -> IntegrationParams:
        integration = self._get_integration_for_app(app=app, auth_scheme=auth_scheme)
        return {
            "integration_id": integration.id,
            "auth_scheme": integration.authScheme,
            "expected_params": integration.expectedInputFields,
        }

    def _can_use_auth_scheme_without_user_input(
        self, scheme: AppAuthScheme, app: AppModel
    ) -> bool:
        if (
            scheme.auth_mode in ("OAUTH2", "OAUTH1")
            and len(app.testConnectors or []) > 0
        ):
            return True

        for field in scheme.fields:
            if not field.expected_from_customer and field.required:
                return False

        return True

    def get_expected_params_for_user(
        self,
        app: t.Optional[AppType] = None,
        auth_scheme: t.Optional[AuthSchemeType] = None,
        integration_id: t.Optional[str] = None,
    ) -> IntegrationParams:
        """
        This method returns a list of parameters that are suppossed to be
        provided by the user.
        """
        # If `integration_id` is provided, use it to fetch the params
        if integration_id is not None:
            response = self._get_expected_params_from_integration_id(id=integration_id)
            if auth_scheme is not None and response["auth_scheme"] != auth_scheme:
                raise ComposioSDKError(
                    message=(
                        "Auth scheme does not match provided integration ID, "
                        f"auth scheme associated with integration ID {response['auth_scheme']} "
                        f"auth scheme provided {auth_scheme}"
                    )
                )
            return response

        if app is None:
            raise ComposioSDKError(
                message="Both `integration_id` and `app` cannot be None"
            )

        try:
            # Check if integration is available for an app, and if available
            # return params from that integration
            return self._get_expected_params_from_app(app=app, auth_scheme=auth_scheme)
        except ValueError:
            pass

        app_data = self.client.apps.get(name=str(app))
        # Go through available schemes and check if any scheme can be used
        # without user inputs to create an integratuib, if yes then create
        # an integration and return params from there.
        for scheme in app_data.auth_schemes or []:
            if auth_scheme is not None and auth_scheme != scheme.auth_mode.upper():
                continue
            if self._can_use_auth_scheme_without_user_input(
                scheme=scheme, app=app_data
            ):
                integration = self.create_integration(
                    app=app,
                    auth_mode=scheme.auth_mode,
                    auth_config={},
                    use_composio_oauth_app=scheme.auth_mode in ("OAUTH2", "OAUTH1"),
                )
                return {
                    "integration_id": integration.id,
                    "auth_scheme": integration.authScheme,
                    "expected_params": integration.expectedInputFields,
                }

        raise ComposioSDKError(
            message=(
                f"No existing integration found for `{str(app)}`, with auth "
                f"scheme {auth_scheme} Please create an integration and use the"
                " ID to fetch the expected params."
            )
        )

    def fetch_expected_integration_params(
        self,
        app: AppModel,
        auth_scheme: AuthSchemeType,
    ) -> t.List[AuthSchemeField]:
        """Fetch expected integration params for creating an integration."""
        for scheme in app.auth_schemes or []:
            if auth_scheme != scheme.auth_mode.upper():
                continue
            return [f for f in scheme.fields if not f.expected_from_customer]
        raise ComposioSDKError(
            message=f"{app.name!r} does not support {auth_scheme!r} auth scheme"
        )

    def create_integration(
        self,
        app: AppType,
        auth_mode: t.Optional[AuthSchemeType] = None,
        auth_config: t.Optional[t.Dict[str, t.Any]] = None,
        use_composio_oauth_app: bool = True,
        force_new_integration: bool = False,
    ) -> IntegrationModel:
        app_data = self.client.apps.get(name=str(app))
        timestamp = datetime.now().strftime("%Y%m%d%H%M%S")
        return self.client.integrations.create(
            app_id=app_data.appId,
            name=f"{app}_{timestamp}",
            auth_mode=auth_mode,
            auth_config=auth_config,
            use_composio_auth=use_composio_oauth_app,
            force_new_integration=force_new_integration,
        )

    def initiate_connection(
        self,
        integration_id: t.Optional[str] = None,
        app: t.Optional[AppType] = None,
        labels: t.Optional[t.List] = None,
        entity_id: t.Optional[str] = None,
        redirect_url: t.Optional[str] = None,
        connected_account_params: t.Optional[t.Dict] = None,
        *,
        auth_scheme: t.Optional[AuthSchemeType] = None,
        auth_config: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> ConnectionRequestModel:
        """
        Initiates a connection with the specified integration.

        If no integration exists for the given app, automatically creates one.

        :param integration_id: (Optional[str]): ID of the existing integration
        :param app: (Optional[AppType]): App to create/use integration for
        :param redirect_url: (Optional[str]) Redirect URL for the App
        :param auth_scheme: (Optional[AuthSchemeType]): Authentication scheme to use
        :return: (ConnectionRequestModel) Details of the connection request.
        """
        if auth_scheme is not None and auth_scheme not in AUTH_SCHEMES:
            raise ComposioSDKError(f"'auth_scheme' must be one of {AUTH_SCHEMES}")

        if integration_id is None:
            if app is None:
                raise ComposioSDKError(
                    message="Both `integration_id` and `app` cannot be None"
                )

            if auth_scheme is None:
                auth_scheme = self.get_auth_scheme_for_app(app).auth_mode

            try:
                integration_id = self._get_integration_for_app(
                    app=t.cast(
                        AppType,
                        app,
                    ),
                    auth_scheme=auth_scheme,
                ).id
            except NoItemsFound:
                auth_config, use_composio_auth = self._validate_auth_config(
                    app, auth_scheme, auth_config
                )
                integration = self.create_integration(
                    app=app,
                    auth_mode=auth_scheme,
                    auth_config=auth_config,
                    use_composio_oauth_app=use_composio_auth,
                )
                integration_id = integration.id

        return self.client.connected_accounts.initiate(
            integration_id=integration_id,
            entity_id=entity_id or self.entity_id,
            params=connected_account_params,
            labels=labels,
            redirect_url=redirect_url,
        )

    def _validate_auth_config(
        self,
        app: AppType,
        auth_scheme: AuthSchemeType,
        auth_config: t.Optional[t.Dict[str, t.Any]],
    ) -> t.Tuple[t.Optional[t.Dict[str, t.Any]], bool]:
        app_data = self.client.apps.get(name=str(app))
        if auth_config is None and app_data.testConnectors:
            # If we have connectors, can fallback to composio's own auth
            return None, True

        if auth_config is None:
            auth_config = {}

        auth_fields = self.fetch_expected_integration_params(
            app=app_data, auth_scheme=auth_scheme
        )
        # Populate defaults in the auth_config
        for field in auth_fields:
            if field.default is not None and field.name not in auth_config:
                auth_config[field.name] = field.default

        required_fields = [field for field in auth_fields if field.required]
        unavailable_fields = [
            field.name for field in required_fields if field.name not in auth_config
        ]
        if unavailable_fields:
            raise ComposioSDKError(
                f"Expected 'auth_config' to provide these fields: {unavailable_fields}"
            ) from None

        return auth_config, False


def _write_file(file_path: t.Union[str, os.PathLike], content: t.Union[str, bytes]):
    """Write content to a file."""
    if isinstance(content, str):
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(content)
    else:
        with open(file_path, "wb") as file:
            file.write(content)



"""
Composio exceptions.
"""

import typing as t

from composio.constants import ENV_COMPOSIO_API_KEY


class ComposioSDKError(Exception):
    """Base composio SDK error."""

    def __init__(
        self,
        message: str,
        *args: t.Any,
        delegate: bool = False,
    ) -> None:
        """
        Initialize Composio SDK error.

        :param message: Error message
        :param delegate: Whether to delegate the error message to the log
                        collection server or not
        """
        super().__init__(message, *args)
        self.message = message
        self.delegate = delegate


class ApiKeyNotProvidedError(ComposioSDKError):
    """Raise when API key is required but not provided."""

    def __init__(self) -> None:
        super().__init__(
            message=(
                "API Key not provided, either provide API key "
                f"or export it as `{ENV_COMPOSIO_API_KEY}` "
                "or run `composio login`"
            ),
        )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateTeamRequest(BaseModel):
    """Request schema for `CreateTeam`"""

    team_id: int = Field(
        ...,
        alias="team_id",
        description="Team ID (Workspace)",
    )
    name: str = Field(
        default=...,
        alias="name",
        description="Name",
    )
    members: t.List[int] = Field(
        default=...,
        alias="members",
        description="",
    )


class CreateTeamResponse(BaseModel):
    """Response schema for `CreateTeam`"""

    data: t.Dict[str, t.Any]


class CreateTeam(OpenAPIAction):
    """
    The endpoint allows creating Teams within Workspaces, assigning tasks, and
    managing IDs for both. Adding a view-only guest as a paid member may result
    in extra charges due to automatic seat addition.
    """

    _tags = ["Teams - User Groups"]
    _display_name = "create_team"
    _request_schema = CreateTeamRequest
    _response_schema = CreateTeamResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/group"
    method = "post"
    operation_id = "TeamsUserGroups_createTeam"
    action_identifier = "/team/{team_id}/group_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {"name": {"__alias": "name"}, "members": {"__alias": "members"}}

    aliases = {}


class teams_user_groups_create_team(OpenAPIAction):
    """
    The endpoint allows creating Teams within Workspaces, assigning tasks, and
    managing IDs for both. Adding a view-only guest as a paid member may result
    in extra charges due to automatic seat addition.<<DEPRECATED use create_team>>
    """

    _tags = ["Teams - User Groups"]
    _display_name = "teams_user_groups_create_team"
    _request_schema = CreateTeamRequest
    _response_schema = CreateTeamResponse

    url = "https://api.clickup.com/api/v2"
    path = "/team/{team_id}/group"
    method = "post"
    operation_id = "TeamsUserGroups_createTeam"
    action_identifier = "/team/{team_id}/group_post"

    path_params = {"team_id": "team_id"}
    query_params = {}
    header_params = {}
    request_params = {"name": {"__alias": "name"}, "members": {"__alias": "members"}}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateFolderlessListRequest(BaseModel):
    """Request schema for `CreateFolderlessList`"""

    space_id: int = Field(
        ...,
        alias="space_id",
        description="",
    )
    name: str = Field(
        default=...,
        alias="name",
        description="Name",
    )
    content: str = Field(
        default=None,
        alias="content",
        description="Content",
    )
    due_date: int = Field(
        default=None,
        alias="due_date",
        description="Due Date",
    )
    due_date_time: bool = Field(
        default=None,
        alias="due_date_time",
        description="Due Date Time",
    )
    priority: int = Field(
        default=None,
        alias="priority",
        description="Priority",
    )
    assignee: int = Field(
        default=None,
        alias="assignee",
        description="Include a `user_id` to add a List owner.",
    )
    status: str = Field(
        default=None,
        alias="status",
        description=(
            "**Status** refers to the List color rather than the task Statuses available "
            "in the List. "
        ),
    )


class CreateFolderlessListResponse(BaseModel):
    """Response schema for `CreateFolderlessList`"""

    data: t.Dict[str, t.Any]


class CreateFolderlessList(OpenAPIAction):
    """Add a new List in a Space."""

    _tags = ["Lists"]
    _display_name = "create_folderless_list"
    _request_schema = CreateFolderlessListRequest
    _response_schema = CreateFolderlessListResponse

    url = "https://api.clickup.com/api/v2"
    path = "/space/{space_id}/list"
    method = "post"
    operation_id = "Lists_createFolderlessList"
    action_identifier = "/space/{space_id}/list_post"

    path_params = {"space_id": "space_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "name": {"__alias": "name"},
        "content": {"__alias": "content"},
        "due_date": {"__alias": "due_date"},
        "due_date_time": {"__alias": "due_date_time"},
        "priority": {"__alias": "priority"},
        "assignee": {"__alias": "assignee"},
        "status": {"__alias": "status"},
    }

    aliases = {}



import typing as t
from pathlib import Path

from pydantic import Field

from composio.tools.base.exceptions import ExecutionFailed
from composio.tools.base.local import LocalAction
from composio.tools.local.filetool.actions.base_action import (
    BaseFileRequest,
    BaseFileResponse,
    include_cwd,
)


def git_apply_cmd(patch_path: str) -> str:
    """Commands to apply a patch."""
    return f"git apply --verbose {patch_path}"


class ApplyPatchRequest(BaseFileRequest):
    """Request to get a Git patch."""

    patch: str = Field("", description="The patch to apply")


class ApplyPatchResponse(BaseFileResponse):
    """Response to applying a patch."""

    message: str = Field("", description="The message from applying the patch")


class ApplyPatch(LocalAction[ApplyPatchRequest, ApplyPatchResponse]):
    """
    Apply a Git patch to the current working directory and perform lint checks.
    The patch should be in the format of a proper diff.

    Usage example:
    patch = "diff --git a/astropy/modeling/separable.py b/astropy/modeling/separable.py\n--- a/astropy/modeling/separable.py\n+++ b/astropy/modeling/separable.py\n@@ -242,7 +242,7 @@ def _cstack(left, right):\n         cright = _coord_matrix(right, 'right', noutp)\n     else:\n         cright = np.zeros((noutp, right.shape[1]))\n-        cright[-right.shape[0]:, -right.shape[1]:] = 1\n+        cright[-right.shape[0]:, -right.shape[1]:] = right\n \n     return np.hstack([cleft, cright])\n \n"

    This class is responsible for:
    1. Applying the patch using Git.
    2. Running lint checks on the affected files before and after applying the patch.
    3. Reverting changes if new lint errors are introduced.
    """  # noqa: E501

    display_name = "Apply Patch"
    _request_schema = ApplyPatchRequest
    _response_schema = ApplyPatchResponse

    @include_cwd  # type: ignore
    def execute(
        self,
        request: ApplyPatchRequest,
        metadata: t.Dict,
    ) -> ApplyPatchResponse:
        # Check if we're in a git repository or in a subdirectory of one
        file_manager = self.filemanagers.get(request.file_manager_id)
        git_root = self._find_git_root(file_manager.current_dir())
        if not git_root:
            raise ExecutionFailed("Not in a git repository or its subdirectories")

        with open(git_root / "patch.patch", "w", encoding="utf-8") as f:
            f.write(request.patch)

        files_to_be_modified, _ = self._get_files_from_patch(request.patch)
        before_lint, before_file_contents = self._run_lint_on_files(
            file_manager=file_manager,
            files=files_to_be_modified,
        )
        _, error = file_manager.execute_command(
            command=git_apply_cmd(
                patch_path=str(git_root / "patch.patch"),
            )
        )

        if error:
            raise ExecutionFailed(
                f"No Update, found error during applying patch: {error}"
            )

        after_lint, _ = self._run_lint_on_files(file_manager, files_to_be_modified)
        for key, value in before_lint.items():
            file = file_manager.open(path=key)
            new_lint_errors = (
                file._compare_lint_results(  # pylint: disable=protected-access
                    value,
                    after_lint[key],
                )
            )
            if len(new_lint_errors) > 0:
                formatted_errors = (
                    file._format_lint_errors(  # pylint: disable=protected-access
                        new_lint_errors,
                    )
                )
                file.path.write_text(before_file_contents[key], encoding="utf-8")
                raise ExecutionFailed(
                    f"No Update, found error during applying patch, no update made in the file: {formatted_errors}"
                )

        return ApplyPatchResponse(
            message="Successfully applied patch, lint checks passed"
        )

    def _find_git_root(self, path: str) -> t.Optional[Path]:
        """Find the root of the git repository."""
        current = Path(path).resolve()
        while current != current.parent:
            if (current / ".git").is_dir():
                return current
            current = current.parent
        return None

    def _get_files_from_patch(
        self, patch_content: str
    ) -> t.Tuple[t.List[str], t.Dict[str, t.Tuple[int, int]]]:
        """Extract the list of files that will be modified by the patch and their line ranges."""
        files = []
        line_ranges = {}
        current_file = None
        start_line = None
        for line in patch_content.splitlines():
            if line.startswith("+++") or line.startswith("---"):
                file_path = line.split()[1]
                if file_path != "/dev/null":
                    if file_path.startswith(("a/", "b/")):
                        file_path = file_path[2:]
                    files.append(file_path)
                    current_file = file_path
            elif line.startswith("@@"):
                if current_file:
                    line_info = line.split()[1].split(",")[0]
                    start_line = int(line_info.split("-")[1])
                    if current_file not in line_ranges:
                        line_ranges[current_file] = (start_line, start_line)
                    else:
                        line_ranges[current_file] = (
                            min(line_ranges[current_file][0], start_line),
                            max(line_ranges[current_file][1], start_line),
                        )
        return list(set(files)), line_ranges

    def _run_lint_on_files(
        self, file_manager, files: t.List[str]
    ) -> t.Tuple[t.Dict[str, t.List[str]], t.Dict[str, str]]:
        """Run lint on the given files."""
        lint_results = {}
        file_contents = {}
        for file_path in files:
            if file_path.endswith(".py"):
                file = file_manager.open(path=file_path)
                lint_results[file_path] = file.lint()
                file_contents[file_path] = file.path.read_text(encoding="utf-8")
        return lint_results, file_contents



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class EditChecklistRequest(BaseModel):
    """Request schema for `EditChecklist`"""

    checklist_id: str = Field(
        ...,
        alias="checklist_id",
        description="b8a8-48d8-a0c6-b4200788a683 (uuid)",
    )
    name: str = Field(
        default=None,
        alias="name",
        description="Name",
    )
    position: int = Field(
        default=None,
        alias="position",
        description=(
            "Position refers to the order of appearance of checklists on a task.   To "
            "set a checklist to appear at the top of the checklists section of a task, "
            'use `"position": 0`. '
        ),
    )


class EditChecklistResponse(BaseModel):
    """Response schema for `EditChecklist`"""

    data: t.Dict[str, t.Any]


class EditChecklist(OpenAPIAction):
    """
    Rename a task checklist, or reorder a checklist so it appears above or below
    other checklists on a task.
    """

    _tags = ["Task Checklists"]
    _display_name = "edit_checklist"
    _request_schema = EditChecklistRequest
    _response_schema = EditChecklistResponse

    url = "https://api.clickup.com/api/v2"
    path = "/checklist/{checklist_id}"
    method = "put"
    operation_id = "TaskChecklists_updateChecklist"
    action_identifier = "/checklist/{checklist_id}_put"

    path_params = {"checklist_id": "checklist_id"}
    query_params = {}
    header_params = {}
    request_params = {"name": {"__alias": "name"}, "position": {"__alias": "position"}}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class RemoveGuestFromFolderRequest(BaseModel):
    """Request schema for `RemoveGuestFromFolder`"""

    folder_id: int = Field(
        ...,
        alias="folder_id",
        description="",
    )
    guest_id: int = Field(
        ...,
        alias="guest_id",
        description="",
    )
    include_shared: bool = Field(
        default=None,
        alias="include_shared",
        description=(
            "Exclude details of items shared with the guest by setting this parameter "
            "to `false`. By default this parameter is set to `true`. "
        ),
    )


class RemoveGuestFromFolderResponse(BaseModel):
    """Response schema for `RemoveGuestFromFolder`"""

    data: t.Dict[str, t.Any]


class RemoveGuestFromFolder(OpenAPIAction):
    """
    Revoke a guest's access to a Folder.    ***Note:** This endpoint is only
    available to Workspaces on our [Enterprise Plan](https://clickup.com/pricing).*
    """

    _tags = ["Guests"]
    _display_name = "remove_guest_from_folder"
    _request_schema = RemoveGuestFromFolderRequest
    _response_schema = RemoveGuestFromFolderResponse

    url = "https://api.clickup.com/api/v2"
    path = "/folder/{folder_id}/guest/{guest_id}"
    method = "delete"
    operation_id = "Guests_revokeAccessFromFolder"
    action_identifier = "/folder/{folder_id}/guest/{guest_id}_delete"

    path_params = {"folder_id": "folder_id", "guest_id": "guest_id"}
    query_params = {"include_shared": "include_shared"}
    header_params = {}
    request_params = {}

    aliases = {}



import os
from typing import Dict, Optional

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction
from composio.tools.local.codeanalysis.actions.base_action import MethodAnalysisAction
from composio.tools.local.codeanalysis.actions.create_codemap import (
    CreateCodeMap,
    CreateCodeMapRequest,
)


class GetMethodSignatureRequest(BaseModel):
    class_name: Optional[str] = Field(
        None,
        description="Name of the class containing the target method",
    )
    method_name: str = Field(
        ..., description="Name of the method whose signature is to be retrieved"
    )


class GetMethodSignatureResponse(BaseModel):
    result: str = Field(
        ...,
        description="Retrieved method signature as a string, including return type and parameters",
    )


class GetMethodSignature(
    LocalAction[GetMethodSignatureRequest, GetMethodSignatureResponse],
    MethodAnalysisAction,
):
    """
    This tool retrieves the signature of a specified method from a given repository.

    Use this action when you need to Obtain the signature of a specific method without its implementation.

    This action can retrieve the method signature in two scenarios:
    1. If a class name is provided, it retrieves the method from within that class.
    Usage example:
    class_name: Field
    method_name: run_validators
    2. If no class name is provided, it retrieves the method from the global scope.
    Usage example:
    method_name: run_validators
    """

    display_name = "Get Method Signature"
    _tags = ["index"]

    def execute(
        self, request: GetMethodSignatureRequest, metadata: Dict
    ) -> GetMethodSignatureResponse:
        CreateCodeMap().execute(CreateCodeMapRequest(), metadata)
        repo_name = os.path.basename(metadata["dir_to_index_path"])

        self.load_fqdn_cache(repo_name)
        method_artefacts = self.get_method_artefacts(
            query_class_name=request.class_name,
            query_method_name=request.method_name,
            repo_path=metadata["dir_to_index_path"],
        )
        return GetMethodSignatureResponse(result=method_artefacts["signature_ans"])



"""E2B Workspace."""

import os
import time
import typing as t
from dataclasses import dataclass
from uuid import uuid4

from composio.exceptions import ComposioSDKError
from composio.tools.env.base import RemoteWorkspace, WorkspaceConfigType


try:
    import requests
    from e2b import Sandbox

    E2B_INSTALLED = True
except ImportError:
    Sandbox = t.Any
    E2B_INSTALLED = False


TOOLSERVER_PORT = 8000
TOOLSERVER_URL = "https://{host}/api"

DEFAULT_TEMPLATE = "2h9ws7lsk32jyow50lqz"
ENV_E2B_TEMPLATE = "E2B_TEMPLATE"


@dataclass
class Config(WorkspaceConfigType):
    """Host configuration type."""

    template: t.Optional[str] = None
    """Template ID for creating the sandbox, if not provided the composio tooling server template will be used."""

    api_key: t.Optional[str] = None
    """E2B API Key."""

    port: t.Optional[int] = None
    """Port for launching the toolserver on the E2B sandbox."""


class E2BWorkspace(RemoteWorkspace):
    """Create and manage E2B workspace."""

    sandbox: Sandbox

    def __init__(self, config: Config):
        """Initialize E2B workspace."""
        if not E2B_INSTALLED:
            raise ComposioSDKError(
                "`e2b` is required to use e2b workspace, "
                "run `pip3 install composio-core[e2b]` or "
                "`pip3 install e2b e2b-code-interpreter` to install",
            )

        super().__init__(config=config)
        template = config.template
        if template is None:
            template = os.environ.get(ENV_E2B_TEMPLATE)
            if template is not None:
                self.logger.debug(f"Using E2B template `{template}` from environment")
            template = template or DEFAULT_TEMPLATE

        self.template = template
        self.api_key = config.api_key
        self.port = config.port or TOOLSERVER_PORT

    def _wait(self) -> None:
        deadline = time.time() + float(os.environ.get("WORKSPACE_WAIT_TIMEOUT", 60.0))
        while time.time() < deadline:
            try:
                if (
                    self._request(endpoint="", method="get", log=False).status_code
                    == 200
                ):
                    return
            except requests.ConnectionError:
                time.sleep(1)

    def setup(self) -> None:
        """Start toolserver."""
        # Start sandbox
        self.sandbox = Sandbox(template=self.template, api_key=self.api_key)
        self.url = TOOLSERVER_URL.format(host=self.sandbox.get_host(self.port))
        self.logger.debug(f"{self}.url = {self.url}")

        # Start app update in background
        self.sandbox.commands.run(
            cmd="composio apps update",
            envs=self.environment,
            background=False,
        )

        # Setup SSH server
        _ssh_password = uuid4().hex.replace("-", "")
        self.sandbox.commands.run(
            cmd=f"echo user:{_ssh_password} | sudo chpasswd",
            envs=self.environment,
            background=False,
        )
        self.sandbox.commands.run(
            cmd="sudo service ssh restart",
            envs=self.environment,
            background=False,
        )
        self.sandbox.commands.run(
            cmd=(
                f"COMPOSIO_LOGGING_LEVEL=debug "
                f"_SSH_USERNAME=user _SSH_PASSWORD={_ssh_password} "
                f"composio serve -h '0.0.0.0' -p {self.port}"
            ),
            envs=self.environment,
            background=True,
        )
        self.host = self.sandbox.get_host(port=80)
        self.ports = []
        self._wait()

    def teardown(self) -> None:
        """Teardown E2B workspace."""
        super().teardown()
        if hasattr(self, "sandbox"):
            self.sandbox.kill()



import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import CreateImageVectorStore, QueryImageVectorStore


class EmbedTool(LocalTool, autoload=True):
    """
    This tool is useful in embedding images and finding images with text
    """

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/embedtool.png"

    @classmethod
    def actions(cls) -> t.List[t.Type[LocalAction]]:
        return [CreateImageVectorStore, QueryImageVectorStore]



import typing as t

from composio.tools.base.local import LocalAction, LocalTool
from composio.tools.local.shelltool.git_cmds.actions import (
    GetPatchCmd,
    GitRepoTree,
    GithubCloneCmd,
)


class Git(LocalTool, autoload=True):
    """Command manager tool for workspace"""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/shelltool.png"

    @classmethod
    def actions(cls) -> t.List[t.Type[LocalAction]]:
        return [
            GitRepoTree,
            GithubCloneCmd,
            GetPatchCmd,
        ]



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class UpdateWebhookRequest(BaseModel):
    """Request schema for `UpdateWebhook`"""

    webhook_id: str = Field(
        ...,
        alias="webhook_id",
        description="e506-4a29-9d42-26e504e3435e (uuid)",
    )
    endpoint: str = Field(
        default=...,
        alias="endpoint",
        description="Endpoint",
    )
    events: str = Field(
        default=...,
        alias="events",
        description="Events",
    )
    status: str = Field(
        default=...,
        alias="status",
        description="Status",
    )


class UpdateWebhookResponse(BaseModel):
    """Response schema for `UpdateWebhook`"""

    data: t.Dict[str, t.Any]


class UpdateWebhook(OpenAPIAction):
    """Update a webhook to change the events to be monitored."""

    _tags = ["Webhooks"]
    _display_name = "update_webhook"
    _request_schema = UpdateWebhookRequest
    _response_schema = UpdateWebhookResponse

    url = "https://api.clickup.com/api/v2"
    path = "/webhook/{webhook_id}"
    method = "put"
    operation_id = "Webhooks_updateEventsToMonitor"
    action_identifier = "/webhook/{webhook_id}_put"

    path_params = {"webhook_id": "webhook_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "endpoint": {"__alias": "endpoint"},
        "events": {"__alias": "events"},
        "status": {"__alias": "status"},
    }

    aliases = {}



"""Tool for executing shell commands."""

import typing as t

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction
from composio.tools.env.constants import EXIT_CODE, STDERR, STDOUT


class ShellRequest(BaseModel):
    """Shell request abstraction."""

    shell_id: str = Field(
        default="",
        description=(
            "ID of the shell where this command will be executed, if not "
            "provided the recent shell will be used to execute the action"
        ),
    )


class ShellExecRequest(ShellRequest):
    """Execute command request."""

    cmd: str = Field(
        ...,
        description="Command to be executed.",
    )


class ShellExecResponse(BaseModel):
    """Shell execution response."""

    stdout: str = Field(
        ...,
        description="Output captured from the execution of the command",
    )
    stderr: str = Field(
        ...,
        description="Errors captured during execution of the command",
    )
    exit_code: int = Field(
        ...,
        description="Exit code of the command",
    )
    current_shell_pwd: str = Field(
        default="",
        description="Current shell's working directory",
    )


class ExecCommand(LocalAction[ShellExecRequest, ShellExecResponse]):
    """
    Run any command directly on shell.
    Examples:
      1. If you want to run python script, use this tool to run the python
        script. *NOTE* : while running a script, give complete path of the script.
      2. Or if you want to `ls -a` use this tool to run the command.
      3. Or if you want to `cd` to a directory, use this tool to run the command.

    You should only include a SINGLE command in the command section and then
    wait for a response from the shell before continuing with more discussion
    and commands.

    You're free to use any other bash commands you want (e.g. find, grep, cat,
    ls, cd) in addition to the special commands listed above. However, the
    environment does NOT support interactive session commands (e.g. python,
    vim), so please do not invoke them. Never issue a find command against "/"
    directory. It will not work. Always try to find files within the base
    directory given in the task.
    """

    _tags = ["workspace", "shell"]

    def execute(self, request: ShellExecRequest, metadata: t.Dict) -> ShellExecResponse:
        """Execute a shell command."""
        shell = self.shells.get(id=request.shell_id)
        self.logger.debug(f"Executing {request.cmd} @ {shell}")

        output = shell.exec(cmd=request.cmd)
        self.logger.debug(output)
        return ShellExecResponse(
            stdout=output[STDOUT],
            stderr=output[STDERR],
            exit_code=int(output[EXIT_CODE]),
            current_shell_pwd=f"Currently in {shell.exec(cmd='pwd')[STDOUT].strip()}",
        )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetListsRequest(BaseModel):
    """Request schema for `GetLists`"""

    folder_id: int = Field(
        ...,
        alias="folder_id",
        description="",
    )
    archived: bool = Field(
        default=None,
        alias="archived",
        description="",
    )


class GetListsResponse(BaseModel):
    """Response schema for `GetLists`"""

    data: t.Dict[str, t.Any]


class GetLists(OpenAPIAction):
    """View the Lists within a Folder."""

    _tags = ["Lists"]
    _display_name = "get_lists"
    _request_schema = GetListsRequest
    _response_schema = GetListsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/folder/{folder_id}/list"
    method = "get"
    operation_id = "Lists_getFolderLists"
    action_identifier = "/folder/{folder_id}/list_get"

    path_params = {"folder_id": "folder_id"}
    query_params = {"archived": "archived"}
    header_params = {}
    request_params = {}

    aliases = {}


class lists_get_folder_lists(OpenAPIAction):
    """View the Lists within a Folder.<<DEPRECATED use get_lists>>"""

    _tags = ["Lists"]
    _display_name = "lists_get_folder_lists"
    _request_schema = GetListsRequest
    _response_schema = GetListsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/folder/{folder_id}/list"
    method = "get"
    operation_id = "Lists_getFolderLists"
    action_identifier = "/folder/{folder_id}/list_get"

    path_params = {"folder_id": "folder_id"}
    query_params = {"archived": "archived"}
    header_params = {}
    request_params = {}

    aliases = {}



import json
import os
from typing import Dict

import requests
from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


class message(BaseModel):
    id: str = Field(..., description="The id of the message")
    content: str = Field(..., description="The content of the message")
    role: str = Field(
        default="user",
        description="The role of the message. ex system,user",
        examples=["system", "user"],
    )


class CodeQueryRequest(BaseModel):
    question: str = Field(
        ...,
        description="The question to ask the mentor",
    )
    sessionId: str = Field(
        default=None,
        description="The session id of the conversation, if you want to continue the conversation with the same mentor. defaults to None",
        examples=["1234567890"],
    )
    genius: bool = Field(
        default=False,
        description="When set to true, multiple mentors will be asked and the answers be really good. Default is false",
        examples=[True, False],
    )
    repository: str = Field(
        ...,
        description="The repository to ask the question about. This should be a github repository. Example openai/docs, composiohq/composio",
        examples=["openai/docs", "composiohq/composio"],
    )
    timeout: int = Field(
        default=60,
        description="The timeout for the Greptile API request. Default is 20 seconds",
        examples=[60, 120, 180],
    )
    branch: str = Field(
        default=None,
        description="The branch to ask the question about. Default is master, if not specified. Example: main, master",
        examples=["master", "main"],
    )


class CodeQueryResponse(BaseModel):
    response: dict = Field(..., description="The response to the question")


class CodeQuery(LocalAction[CodeQueryRequest, CodeQueryResponse]):
    """
    Ask the mentor, any questions on the code and get the answer from the mentor.
    with a list of relevant code references (filepaths, line numbers, etc)
    If it's first question
    examples:
    “How does auth work in this codebase?”
    “Generate a description for the JIRA ticket with codebase context”
    “Rewrite this code snippet using relevant abstractions already in the repo”
    """

    _tags = ["code_query"]

    def execute(self, request: CodeQueryRequest, metadata: Dict) -> CodeQueryResponse:
        token = metadata.get("greptile_token", os.getenv("GREPTILE_TOKEN"))
        if token is None:
            self.logger.error("GREPTILE_TOKEN is not set")
            raise ValueError("GREPTILE_TOKEN is not set")

        github_token = metadata.get("github_token", os.getenv("GITHUB_TOKEN"))
        if github_token is None:
            self.logger.error("GITHUB_TOKEN is not set")
            raise ValueError("GITHUB_TOKEN is not set")

        # Construct the headers for the API request
        headers = {
            "Authorization": f"Bearer {token}",
            "Content-Type": "application/json",
            "X-GitHub-Token": github_token,
        }

        # Construct the data payload for the API request
        data = {
            "messages": [
                {"content": request.question, "role": "user"},
            ],
            "repositories": [
                {
                    "remote": "github",
                    "branch": request.branch,
                    "repository": request.repository,
                }
            ],
            "genius": request.genius,
        }

        if request.sessionId and request.sessionId != "":
            data["sessionId"] = request.sessionId

        # Send the POST request to the Greptile API
        response = requests.post(
            "https://api.greptile.com/v2/query",
            headers=headers,
            data=json.dumps(data),
            timeout=request.timeout,
        )

        # Check if the request was successful
        if response.status_code == 200:
            return CodeQueryResponse(response=response.json())

        self.logger.error(
            "Failed to fetch data from Greptile API, status code: %s",
            response.status_code,
        )

        raise ValueError(
            f"Failed to fetch data from Greptile API with status code "
            f"{response.status_code} and error {response.content.decode()}"
        )



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class EditKeyResultRequest(BaseModel):
    """Request schema for `EditKeyResult`"""

    key_result_id: str = Field(
        ...,
        alias="key_result_id",
        description="8480-49bc-8c57-e569747efe93 (uuid)",
    )
    steps_current: int = Field(
        default=...,
        alias="steps_current",
        description="Steps Current",
    )
    note: str = Field(
        default=...,
        alias="note",
        description="Note",
    )


class EditKeyResultResponse(BaseModel):
    """Response schema for `EditKeyResult`"""

    data: t.Dict[str, t.Any]


class EditKeyResult(OpenAPIAction):
    """Update a Target."""

    _tags = ["Goals"]
    _display_name = "edit_key_result"
    _request_schema = EditKeyResultRequest
    _response_schema = EditKeyResultResponse

    url = "https://api.clickup.com/api/v2"
    path = "/key_result/{key_result_id}"
    method = "put"
    operation_id = "Goals_updateKeyResult"
    action_identifier = "/key_result/{key_result_id}_put"

    path_params = {"key_result_id": "key_result_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "steps_current": {"__alias": "steps_current"},
        "note": {"__alias": "note"},
    }

    aliases = {}



import typing as t

from composio.tools.base.local import LocalAction, LocalTool

from .actions import ScrapeWebsiteContent, ScrapeWebsiteElement


class Webtool(LocalTool, autoload=True):
    """Web Tools"""

    logo = "https://raw.githubusercontent.com/ComposioHQ/composio/master/python/docs/imgs/logos/webtool.png"

    @classmethod
    def actions(cls) -> list[t.Type[LocalAction]]:
        return [ScrapeWebsiteContent, ScrapeWebsiteElement]



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetFolderViewsRequest(BaseModel):
    """Request schema for `GetFolderViews`"""

    folder_id: int = Field(
        ...,
        alias="folder_id",
        description="",
    )


class GetFolderViewsResponse(BaseModel):
    """Response schema for `GetFolderViews`"""

    data: t.Dict[str, t.Any]


class GetFolderViews(OpenAPIAction):
    """View the task and page views available for a Folder."""

    _tags = ["Views"]
    _display_name = "get_folder_views"
    _request_schema = GetFolderViewsRequest
    _response_schema = GetFolderViewsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/folder/{folder_id}/view"
    method = "get"
    operation_id = "Views_folderViewsGet"
    action_identifier = "/folder/{folder_id}/view_get"

    path_params = {"folder_id": "folder_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



# pylint: disable=wrong-import-position

from composio.utils import sentry


sentry.init()

import atexit  # noqa: E402

from composio.__version__ import __version__  # noqa: E402
from composio.client import Composio  # noqa: E402
from composio.client.collections import CustomAuthObject  # noqa: E402
from composio.client.enums import (  # noqa: E402
    Action,
    ActionType,
    App,
    AppType,
    Tag,
    TagType,
    Trigger,
    TriggerType,
)
from composio.tools import ComposioToolSet  # noqa: E402
from composio.tools.base.runtime import action  # noqa: E402
from composio.tools.env.factory import (  # noqa: E402
    WorkspaceConfigType,
    WorkspaceFactory,
    WorkspaceType,
)
from composio.tools.env.host.shell import Shell  # noqa: E402
from composio.utils.logging import LogLevel  # noqa: E402
from composio.utils.warnings import create_latest_version_warning_hook  # noqa: E402


__all__ = (
    "Tag",
    "App",
    "Action",
    "AppType",
    "Trigger",
    "TagType",
    "Composio",
    "ActionType",
    "TriggerType",
    "ComposioToolSet",
    "CustomAuthObject",
    "WorkspaceType",
    "WorkspaceConfigType",
    "WorkspaceFactory",
    "Shell",
    "action",
    "LogLevel",
)

atexit.register(create_latest_version_warning_hook(version=__version__))



"""
Helper function for git.
"""

import subprocess
import typing as t

from pydantic import BaseModel


class GitUserInfo(BaseModel):
    """Git user info."""

    name: t.Optional[str]
    email: t.Optional[str]


def get_git_user_info() -> GitUserInfo:
    """Get current git user information."""
    try:
        name = (
            subprocess.check_output(["git", "config", "user.name"])
            .strip()
            .decode("utf-8")
        )
        email = (
            subprocess.check_output(["git", "config", "user.email"])
            .strip()
            .decode("utf-8")
        )
        return GitUserInfo(name=name, email=email)
    except subprocess.CalledProcessError:
        return GitUserInfo(name=None, email=None)



import os
from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction
from composio.tools.local.codeanalysis.actions.create_codemap import (
    CreateCodeMap,
    CreateCodeMapRequest,
)


class GetRelevantCodeRequest(BaseModel):
    query: str = Field(
        ...,
        description="Query to retrieve relevant code from the repository",
    )


class GetRelevantCodeResponse(BaseModel):
    result: str = Field(
        ...,
        description="Retrieved method body as a string, including any decorators and comments",
    )


class GetRelevantCode(LocalAction[GetRelevantCodeRequest, GetRelevantCodeResponse]):
    """
    Retrieves relevant code snippets from a repository based on a given query.

    Use this action when you need to:
    1. Find code snippets related to a specific topic or functionality.
    2. Search for implementations of particular features across the codebase.

    Usage example:
    query: "database connection pooling"

    The relevance of retrieved code snippets depends on the quality and specificity of the provided query.
    Don't use this action if you are not sure about the query. And the results returned are not very relevant.
    """

    display_name = "Get Relevant Code"
    _tags = ["index"]

    def execute(
        self, request: GetRelevantCodeRequest, metadata: Dict
    ) -> GetRelevantCodeResponse:
        CreateCodeMap().execute(CreateCodeMapRequest(), metadata)
        from composio.tools.local.codeanalysis import (  # pylint: disable=import-outside-toplevel
            embedder,
        )

        repo_name = os.path.basename(metadata["dir_to_index_path"])
        vector_store = embedder.get_vector_store(repo_name, overwrite=False)
        query = request.query
        results = embedder.get_topn_chunks_from_query(vector_store, query, top_n=5)
        sep = "\n" + "=" * 100 + "\n"
        result_string = "Query: " + query + sep
        for i, _metadata in enumerate(results["metadata"]):
            result_string += f"Chunk {i + 1}: \n" + str(_metadata["chunk"]) + sep
        return GetRelevantCodeResponse(result=result_string)



"""
Action for getting details of an element on a webpage.
"""

from typing import Any, Dict

from pydantic import Field

from composio.tools.env.browsermanager.manager import BrowserManager
from composio.tools.local.browsertool.actions.base_action import (
    BaseBrowserAction,
    BaseBrowserResponse,
    BaseBrowserSelectorRequest,
)


class GetElementDetailsRequest(BaseBrowserSelectorRequest):
    """Request schema for getting element details."""


class GetElementDetailsResponse(BaseBrowserResponse):
    """Response schema for getting element details."""

    success: bool = Field(
        default=False, description="Whether the action was successful"
    )
    element_found: bool = Field(
        default=False, description="Whether the element was found on the page"
    )
    details: Dict[str, Any] = Field(
        default_factory=dict, description="Details of the found element"
    )


class GetElementDetails(
    BaseBrowserAction[GetElementDetailsRequest, GetElementDetailsResponse]
):
    """
    Get details of an element on the current webpage.

    This action performs the following steps:
    1. Locates the specified element using the provided selector and selector type.
    2. If the element is found, retrieves detailed information about it.

    The details retrieved include:
    - Tag name
    - ID
    - Class name
    - Text content
    & more

    This action is useful for:
    - Debugging element issues
    - Extracting detailed information about specific page elements
    - Verifying element properties and attributes
    """

    display_name = "GetElementDetails"
    _request_schema = GetElementDetailsRequest
    _response_schema = GetElementDetailsResponse

    def execute_on_browser_manager(
        self, browser_manager: BrowserManager, request: GetElementDetailsRequest  # type: ignore
    ) -> GetElementDetailsResponse:
        """Execute the get element details action."""
        # Find the element
        element = browser_manager.find_element(request.selector, request.selector_type)

        if element is None:
            return GetElementDetailsResponse(success=False, element_found=False)

        # Get element details
        details = browser_manager.execute_script(
            """
            function getElementDetails(element) {
                const rect = element.getBoundingClientRect();
                return {
                    tagName: element.tagName.toLowerCase(),
                    id: element.id,
                    className: element.className,
                    text: element.textContent.trim(),
                    value: element.value,
                    isVisible: window.getComputedStyle(element).display !== 'none',
                    attributes: Object.fromEntries([...element.attributes].map(attr => [attr.name, attr.value])),
                    position: {
                        x: rect.left,
                        y: rect.top
                    },
                    size: {
                        width: rect.width,
                        height: rect.height
                    }
                };
            }
            return getElementDetails(arguments[0]);
        """,
            element,
        )

        return GetElementDetailsResponse(
            success=True, element_found=True, details=details
        )



import base64
import os
import typing as t
from abc import ABC, abstractmethod
from enum import Enum
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import requests
from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


system_prompt = "You are an expert assistant that analyzes images and provides detailed descriptions to answer questions about them."


class ModelChoice(str, Enum):
    GPT4_VISION = "gpt-4o"
    GPT4_VISION_MINI = "gpt-4o-mini"
    CLAUDE_3_SONNET = "claude-3-sonnet-20240229"


class ImageAnalyserRequest(BaseModel):
    media_paths: List[str] = Field(
        ..., description="List of local file paths or URLs of the media to be analyzed"
    )
    prompt: str = Field(..., description="Prompt for guiding the analysis")


class ImageAnalyserResponse(BaseModel):
    analysis: Optional[str] = Field(
        None, description="The resulting analysis of the media"
    )
    error_message: Optional[str] = Field(
        None, description="Error message if analysis failed"
    )


class MediaAnalyzer(ABC):
    @abstractmethod
    def analyze(
        self, model: ModelChoice, media_paths: List[str], prompt: str, api_key: str
    ) -> str:
        pass


class OpenAIAnalyzer(MediaAnalyzer):
    def analyze(
        self, model: ModelChoice, media_paths: List[str], prompt: str, api_key: str
    ) -> str:
        try:
            from openai import OpenAI  # pylint: disable=import-outside-toplevel
            from openai.types.chat import (  # pylint: disable=import-outside-toplevel
                ChatCompletionMessageParam,
            )
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "The 'openai' package is required for OpenAI analysis. Please install it using 'pip install openai'."
            ) from e
        client = OpenAI(api_key=api_key)
        content = [{"type": "text", "text": prompt}]

        for media_path in media_paths:
            image_url = (
                media_path
                if media_path.startswith(("http://", "https://"))
                else self._encode_image(media_path)
            )
            content.append(
                {
                    "type": "image_url",
                    "image_url": {"url": image_url, "detail": "high"},  # type: ignore
                }
            )

        messages: List[ChatCompletionMessageParam] = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": content},  # type: ignore
        ]

        try:
            response = client.chat.completions.create(model=model, messages=messages)
            return response.choices[0].message.content or "No analysis provided"
        except Exception as e:
            return f"Error during OpenAI analysis: {str(e)}"

    def _encode_image(self, file_path: str) -> str:
        with open(file_path, "rb") as image_file:
            return f"data:image/jpeg;base64,{base64.b64encode(image_file.read()).decode('utf-8')}"


class ClaudeAnalyzer(MediaAnalyzer):
    def analyze(
        self, model: ModelChoice, media_paths: List[str], prompt: str, api_key: str
    ) -> str:
        try:
            import anthropic  # pylint: disable=import-outside-toplevel
        except ModuleNotFoundError as e:
            raise ModuleNotFoundError(
                "The 'anthropic' package is required for Claude analysis. Please install it using 'pip install anthropic'."
            ) from e
        client = anthropic.Anthropic(api_key=api_key)

        content = []
        for i, media_path in enumerate(media_paths, start=1):
            content.extend(
                [
                    {"type": "text", "text": f"Image {i}:"},
                    self._prepare_image_content(media_path),
                ]
            )

        content.append({"type": "text", "text": prompt})

        try:
            message = client.messages.create(
                model=model,
                max_tokens=1024,
                system=system_prompt,
                messages=[{"role": "user", "content": content}],
            )
            return message.content[0].text  # type: ignore
        except Exception as e:
            return f"Error during Claude analysis: {str(e)}"

    def _prepare_image_content(self, media_path: str) -> Dict[str, Any]:
        if media_path.startswith(("http://", "https://")):
            response = requests.get(media_path, timeout=40)
            image_data = base64.b64encode(response.content).decode("utf-8")
        else:
            with open(media_path, "rb") as image_file:
                image_data = base64.b64encode(image_file.read()).decode("utf-8")

        return {
            "type": "image",
            "source": {
                "type": "base64",
                "media_type": self._get_media_type(media_path),
                "data": image_data,
            },
        }

    def _get_media_type(self, file_path: str) -> str:
        file_extension = Path(file_path).suffix.lower()
        media_types = {
            ".jpg": "image/jpeg",
            ".jpeg": "image/jpeg",
            ".png": "image/png",
            ".gif": "image/gif",
            ".webp": "image/webp",
        }
        if file_extension in media_types:
            return media_types[file_extension]
        return f"Unsupported image format: {file_extension}"


class Analyse(LocalAction[ImageAnalyserRequest, ImageAnalyserResponse]):
    """Analyze local images using multimodal LLMs."""

    _tags = ["image"]

    @property
    def analyzers(self) -> t.Dict[ModelChoice, MediaAnalyzer]:
        return {
            ModelChoice.GPT4_VISION: OpenAIAnalyzer(),
            ModelChoice.GPT4_VISION_MINI: OpenAIAnalyzer(),
            ModelChoice.CLAUDE_3_SONNET: ClaudeAnalyzer(),
        }

    def execute(
        self, request: ImageAnalyserRequest, metadata: Dict
    ) -> ImageAnalyserResponse:
        """Execute image analysis."""
        validation_result = self._validate_request(request)
        if validation_result:
            return ImageAnalyserResponse(error_message=validation_result, analysis=None)

        media_model_choice = os.environ.get("MEDIA_MODEL_CHOICE", "gpt-4o")
        media_model_choice = ModelChoice(media_model_choice)
        api_key_result = self._get_api_key(media_model_choice, metadata)
        if not isinstance(api_key_result, str):
            return ImageAnalyserResponse(
                error_message="No API key found", analysis=None
            )

        analyzer = self.analyzers.get(media_model_choice)
        if not analyzer:
            return ImageAnalyserResponse(
                error_message=f"Unsupported model: {media_model_choice}",
                analysis=None,
            )

        analysis = analyzer.analyze(
            media_model_choice,
            request.media_paths,
            request.prompt,
            api_key_result,
        )
        return ImageAnalyserResponse(analysis=analysis, error_message=None)

    def _validate_request(self, request: ImageAnalyserRequest) -> Optional[str]:
        if not request.media_paths:
            return "Media paths cannot be None or empty"
        for media_path in request.media_paths:
            if not media_path.startswith(("http://", "https://")):
                path = Path(media_path)
                if not path.exists() or not path.is_file():
                    return f"Media file not found: {media_path}"
        return None

    def _get_api_key(
        self, model: ModelChoice, metadata: dict
    ) -> Union[str, Dict[str, str]]:
        if model in [ModelChoice.GPT4_VISION, ModelChoice.GPT4_VISION_MINI]:
            key = os.environ.get("OPENAI_API_KEY") or metadata.get("OPENAI_API_KEY")
            key_name = "OPENAI_API_KEY"

        elif model == ModelChoice.CLAUDE_3_SONNET:
            key = os.environ.get("ANTHROPIC_API_KEY") or metadata.get(
                "ANTHROPIC_API_KEY"
            )
            key_name = "ANTHROPIC_API_KEY"
        else:
            return {"error": f"Unsupported model: {model}"}
        if not key:
            return f"{key_name} not found for {model}"
        return key



import os
from typing import Any, Dict, List

from deeplake.core.vectorstore.deeplake_vectorstore import DeepLakeVectorStore
from sentence_transformers import SentenceTransformer

from composio.tools.local.codeanalysis.constants import (
    CODE_MAP_CACHE,
    DEEPLAKE_FOLDER,
    EMBEDDER,
)


def get_vector_store(repo_name: str, overwrite: bool = True) -> DeepLakeVectorStore:
    """
    Get or create a DeepLakeVectorStore for the given repository.

    Args:
        repo_path (str): Path to the repository.
        overwrite (bool, optional): Whether to overwrite existing vector store. Defaults to False.

    Returns:
        DeepLakeVectorStore: The vector store for the repository.

    Raises:
        ValueError: If repo_path is empty or None.
        OSError: If there's an issue creating or accessing the vector store.
    """
    if not repo_name:
        raise ValueError("Repository path cannot be empty or None")

    try:
        repo_name = os.path.basename(repo_name)

        deeplake_repo_path = os.path.join(CODE_MAP_CACHE, repo_name, DEEPLAKE_FOLDER)

        deeplake_vector_store = DeepLakeVectorStore(
            path=deeplake_repo_path,
            overwrite=overwrite,
            read_only=False,
            ingestion_batch_size=1000,
        )

        return deeplake_vector_store

    except OSError as e:
        raise OSError(f"Failed to create or access vector store: {str(e)}") from e


class Embedding:
    def __init__(self):
        self.model = SentenceTransformer(EMBEDDER)

    def compute(self, texts: List[str]) -> List[List[float]]:
        """
        Compute embeddings for a list of texts.

        Args:
            texts (List[str]): List of input texts to embed.

        Returns:
            List[List[float]]: List of embeddings, where each embedding is a list of floats.

        Raises:
            ValueError: If texts is empty or None.
        """
        if not texts:
            raise ValueError("Input texts cannot be empty or None")

        try:
            embeddings = self.model.encode(
                texts, batch_size=64, show_progress_bar=True, convert_to_numpy=True
            )
            return embeddings.tolist()
        except Exception as e:
            raise RuntimeError(f"Failed to compute embeddings: {str(e)}") from e


def get_vector_store_from_chunks(
    repo_name: str,
    documents: List[str],
    ids: List[str],
    metadatas: List[Dict[str, Any]],
) -> DeepLakeVectorStore:
    """
    Create or update a vector store with the given documents, ids, and metadata.

    Args:
        repo_path (str): Path to the repository.
        documents (List[str]): List of document texts to be embedded.
        ids (List[str]): List of unique identifiers for each document.
        metadatas (List[Dict[str, Any]]): List of metadata dictionaries for each document.

    Returns:
        DeepLakeVectorStore: The updated vector store.

    Raises:
        ValueError: If the input lists have different lengths.
        RuntimeError: If there's an error during vector store creation or embedding.
    """
    if not (len(documents) == len(ids) == len(metadatas)):
        raise ValueError("Input lists must have the same length")

    try:
        vector_store = get_vector_store(repo_name)
        embed_model = Embedding()
        embeddings = embed_model.compute(documents)

        # Batch processing for better performance
        batch_size = 1000
        for i in range(0, len(documents), batch_size):
            batch_end = min(i + batch_size, len(documents))
            vector_store.add(
                text=ids[i:batch_end],
                embedding=embeddings[i:batch_end],
                metadata=metadatas[i:batch_end],
            )

        return vector_store
    except Exception as e:
        raise RuntimeError(
            f"Error in vector store creation or embedding: {str(e)}"
        ) from e


def get_topn_chunks_from_query(
    vector_store: DeepLakeVectorStore, query: str, top_n: int = 5
) -> Dict:
    """
    Retrieve the top N most relevant chunks from the vector store based on the given query.

    Args:
        vector_store (DeepLakeVectorStore): The vector store to search in.
        query (str): The query string to search for.
        top_n (int, optional): The number of top results to return. Defaults to 5.

    Returns:
        Dict[str, Any]: A dictionary containing the search results.

    Raises:
        ValueError: If the input parameters are invalid.
        RuntimeError: If there's an error during the search process.
    """
    if not isinstance(vector_store, DeepLakeVectorStore):
        raise ValueError("vector_store must be an instance of DeepLakeVectorStore")
    try:
        embed_model = Embedding()
        query_embedding = embed_model.compute([query])[0]
        results = vector_store.search(embedding=query_embedding, k=top_n)
        return results
    except Exception as e:
        raise RuntimeError(f"Error during vector store search: {str(e)}") from e



from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction
from composio.tools.local.shelltool.shell_exec.actions.exec import ShellExecRequest
from composio.utils.logging import get as get_logger


STATUS_RUNNING = "running"
STATUS_STOPPED = "stopped"
logger = get_logger("workspace")


class GetWorkspaceHistoryRequest(ShellExecRequest):
    pass


class GetWorkspaceHistoryResponse(BaseModel):
    workspace_history: dict = Field(
        ...,
        description="history of last n commands on the workspace",
    )


class GetWorkspaceHistory(
    LocalAction[
        GetWorkspaceHistoryRequest,
        GetWorkspaceHistoryResponse,
    ]
):
    """
    Returns history for workspace which includes
        - state of the environment
        - last executed n commands
        - output from last n commands
    """

    _tags = ["workspace"]
    _history_len = 5

    def execute(
        self, request: GetWorkspaceHistoryRequest, metadata: Dict
    ) -> GetWorkspaceHistoryResponse:
        return GetWorkspaceHistoryResponse(workspace_history={})



# This file is based on code from
# https://github.com/paul-gauthier/grep-ast/blob/main/grep_ast/grep_ast.py

import os
from pathlib import Path

from composio.tools.local.base.utils.grep_ast import TreeContext
from composio.tools.local.base.utils.parser import filename_to_lang


def get_files_excluding_gitignore(root_path, no_gitignore=False, file_patterns=None):
    """
    Get all files matching the given patterns in the root path, excluding those specified in .gitignore.

    :param root_path: The root directory to start searching from.
    :param no_gitignore: If True, ignore .gitignore file.
    :param file_patterns: A list of file patterns to match. Defaults to ["*.py", "*.md"].
    :return: A list of file paths matching the patterns.
    """
    root_path = Path(root_path).resolve()
    gitignore = None

    if not no_gitignore:
        # Check root_path first
        potential_gitignore = root_path / ".gitignore"
        if potential_gitignore.exists():
            gitignore = potential_gitignore
        else:
            # Then check parent directories
            for parent in root_path.parents:
                potential_gitignore = parent / ".gitignore"
                if potential_gitignore.exists():
                    gitignore = potential_gitignore
                    break

    import pathspec  # TODO: simplify import  # pylint: disable=C0415

    if gitignore:
        with open(gitignore, "r", encoding="utf-8") as f:
            spec = pathspec.PathSpec.from_lines("gitwildmatch", f)
    else:
        spec = pathspec.PathSpec.from_lines("gitwildmatch", [])

    if file_patterns is None:
        file_patterns = ["*.[pP][yY]", "*.[mM][dD]", "*.[rR][sS][tT]", "*.[tT][xX][tT]"]

    files = []
    for pattern in file_patterns:
        for path in root_path.rglob(pattern):
            # Exclude .git and other version control system folders
            if any(part.startswith(".") and part != "." for part in path.parts):
                continue
            if path.is_file() and not spec.match_file(path):
                files.append(str(path))

    return files


# callable utility which works the same way as main.
def grep_util(
    pattern,
    filenames,
    encoding="utf8",
    color=None,
    verbose=False,
    line_number=True,
    ignore_case=True,
    no_gitignore=False,
):
    results = []

    for filename in filenames:
        if os.path.isdir(filename):
            dir_files = get_files_excluding_gitignore(filename, no_gitignore)
            for file in dir_files:
                results.extend(
                    process_file(
                        file,
                        pattern,
                        encoding,
                        ignore_case,
                        color,
                        verbose,
                        line_number,
                    )
                )
        else:
            results.extend(
                process_file(
                    filename,
                    pattern,
                    encoding,
                    ignore_case,
                    color,
                    verbose,
                    line_number,
                )
            )

    return results


def process_file(filename, pattern, encoding, ignore_case, color, verbose, line_number):
    file_results = []
    try:
        with open(filename, "r", encoding=encoding) as f:
            content = f.read()
    except UnicodeDecodeError:
        return file_results

    lang = filename_to_lang(filename)

    if lang:
        try:
            tc = TreeContext(
                filename, content, color=color, verbose=verbose, line_number=line_number
            )
            loi = tc.grep(pattern, ignore_case)
            if loi:
                tc.add_lines_of_interest(loi)
                tc.add_context()
                file_results.append({"filename": filename, "matches": tc.format()})
        except ValueError:
            pass  # Skip files that can't be parsed

    return file_results



"""Local tools."""

import importlib
from pathlib import Path

from composio.tools.base.abs import ToolRegistry, tool_registry


TOOLS_PATH = Path(__file__).parent


def load_local_tools() -> ToolRegistry:
    for tooldef in TOOLS_PATH.glob("**/tool.py"):
        importlib.import_module(
            "composio.tools.local."
            + ".".join(tooldef.relative_to(TOOLS_PATH).parent.parts)
            + ".tool"
        )

    return tool_registry



from typing import Dict

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


# pylint: disable=import-outside-toplevel


class CrawlToolRequest(BaseModel):
    url: str = Field(..., description="Mandatory website url to read the file")


class CrawlToolResponse(BaseModel):
    content: str = Field(..., description="The content of the selected element")


class Crawl(LocalAction[CrawlToolRequest, CrawlToolResponse]):
    """
    Crawl a website
    """

    _tags = ["web", "scrape"]

    def execute(self, request: CrawlToolRequest, metadata: Dict) -> CrawlToolResponse:
        """Crawl a website and return its content in markdown format"""
        try:
            from spider import Spider
            from spider.spider_types import RequestParamsDict

        except ImportError as e:
            raise ImportError("Failed to import Spider:", e) from e

        try:
            spider = Spider()
            params = RequestParamsDict(return_format="markdown")
            response = spider.crawl_url(request.url, params)
            if response is None:
                raise ValueError(f"Error getting response for url {request.url}")
            return CrawlToolResponse(content=response.content)
        except (ConnectionError, TimeoutError) as e:
            raise ValueError(f"Connection or timeout error occurred: {e}") from e
        except Exception as e:
            raise ValueError(f"An unexpected error occurred: {e}") from e



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class AddDependencyRequest(BaseModel):
    """Request schema for `AddDependency`"""

    task_id: str = Field(
        ...,
        alias="task_id",
        description="This is the task which is waiting on or blocking another task.",
    )
    custom_task_ids: bool = Field(
        default=None,
        alias="custom_task_ids",
        description=(
            'If you want to reference a task by it"s custom task id, this value must '
            "be `true`. "
        ),
    )
    team_id: int = Field(
        default=None,
        alias="team_id",
        description=(
            "Only used when the `custom_task_ids` parameter is set to `true`. For example: "
            "`custom_task_ids=true&team_id=123`. "
        ),
    )
    depends_on: str = Field(
        default=None,
        alias="depends_on",
        description="Depends On",
    )
    depedency_of: str = Field(
        default=None,
        alias="depedency_of",
        description="Dependency Of",
    )


class AddDependencyResponse(BaseModel):
    """Response schema for `AddDependency`"""

    data: t.Dict[str, t.Any]


class AddDependency(OpenAPIAction):
    """Set a task as waiting on or blocking another task."""

    _tags = ["Task Relationships"]
    _display_name = "add_dependency"
    _request_schema = AddDependencyRequest
    _response_schema = AddDependencyResponse

    url = "https://api.clickup.com/api/v2"
    path = "/task/{task_id}/dependency"
    method = "post"
    operation_id = "TaskRelationships_addDependency"
    action_identifier = "/task/{task_id}/dependency_post"

    path_params = {"task_id": "task_id"}
    query_params = {"custom_task_ids": "custom_task_ids", "team_id": "team_id"}
    header_params = {}
    request_params = {
        "depends_on": {"__alias": "depends_on"},
        "depedency_of": {"__alias": "depedency_of"},
    }

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class CreateChecklistItemRequest(BaseModel):
    """Request schema for `CreateChecklistItem`"""

    checklist_id: str = Field(
        ...,
        alias="checklist_id",
        description="b8a8-48d8-a0c6-b4200788a683 (uuid)",
    )
    name: str = Field(
        default=None,
        alias="name",
        description="Name",
    )
    assignee: int = Field(
        default=None,
        alias="assignee",
        description="Assignee",
    )


class CreateChecklistItemResponse(BaseModel):
    """Response schema for `CreateChecklistItem`"""

    data: t.Dict[str, t.Any]


class CreateChecklistItem(OpenAPIAction):
    """Add a line item to a task checklist."""

    _tags = ["Task Checklists"]
    _display_name = "create_checklist_item"
    _request_schema = CreateChecklistItemRequest
    _response_schema = CreateChecklistItemResponse

    url = "https://api.clickup.com/api/v2"
    path = "/checklist/{checklist_id}/checklist_item"
    method = "post"
    operation_id = "TaskChecklists_addLineItem"
    action_identifier = "/checklist/{checklist_id}/checklist_item_post"

    path_params = {"checklist_id": "checklist_id"}
    query_params = {}
    header_params = {}
    request_params = {"name": {"__alias": "name"}, "assignee": {"__alias": "assignee"}}

    aliases = {}



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class UpdateGoalRequest(BaseModel):
    """Request schema for `UpdateGoal`"""

    goal_id: str = Field(
        ...,
        alias="goal_id",
        description="900e-462d-a849-4a216b06d930 (uuid)",
    )
    description: str = Field(
        default=...,
        alias="description",
        description="Description",
    )
    name: str = Field(
        default=...,
        alias="name",
        description="Name",
    )
    due_date: int = Field(
        default=...,
        alias="due_date",
        description="Due Date",
    )
    rem_owners: t.List[int] = Field(
        default=...,
        alias="rem_owners",
        description="Array of user IDs.",
    )
    add_owners: t.List[int] = Field(
        default=...,
        alias="add_owners",
        description="Array of user IDs.",
    )
    color: str = Field(
        default=...,
        alias="color",
        description="Color",
    )


class UpdateGoalResponse(BaseModel):
    """Response schema for `UpdateGoal`"""

    data: t.Dict[str, t.Any]


class UpdateGoal(OpenAPIAction):
    """
    Rename a Goal, set the due date, replace the description, add or remove
    owners, and set the Goal color.
    """

    _tags = ["Goals"]
    _display_name = "update_goal"
    _request_schema = UpdateGoalRequest
    _response_schema = UpdateGoalResponse

    url = "https://api.clickup.com/api/v2"
    path = "/goal/{goal_id}"
    method = "put"
    operation_id = "Goals_updateGoalDetails"
    action_identifier = "/goal/{goal_id}_put"

    path_params = {"goal_id": "goal_id"}
    query_params = {}
    header_params = {}
    request_params = {
        "description": {"__alias": "description"},
        "name": {"__alias": "name"},
        "due_date": {"__alias": "due_date"},
        "rem_owners": {"__alias": "rem_owners"},
        "add_owners": {"__alias": "add_owners"},
        "color": {"__alias": "color"},
    }

    aliases = {}



import subprocess
from pathlib import Path
from typing import List, Optional, Type

from pydantic import BaseModel, Field

from composio.tools.base.local import LocalAction


class CodeFormatRequest(BaseModel):
    path: str = Field(
        ...,
        description="The file path or directory path to format and check.",
        examples=["/path/to/file.py", "/path/to/directory"],
    )


class FormatResult(BaseModel):
    file_path: str
    format_changes: List[str]
    errors: List[str]


class CodeFormatResponse(BaseModel):
    results: List[FormatResult] = Field(
        ...,
        description="A list of formatting results, each containing the file path, formatting changes, and errors.",
    )
    error: Optional[str] = Field(
        default=None,
        description="An error message describing any issues encountered during the formatting process. This field is None if the process was successful.",
    )


class FormatAndLintCodebase(LocalAction[CodeFormatRequest, CodeFormatResponse]):
    """
    Performs code formatting and linting using ruff, addressing style issues and checking for errors.

    This action is ideal for:
    - Applying consistent code formatting across Python files
    - Identifying and fixing style issues
    - Detecting potential errors and code quality problems
    - Maintaining a uniform code style across a project or codebase

    The formatting and linting are performed efficiently and can handle both single files and entire directories.

    This action provides a comprehensive tool for code quality improvement and standardization.
    """

    display_name = "Format and Lint Codebase"
    _request_schema: Type[CodeFormatRequest] = CodeFormatRequest
    _response_schema: Type[CodeFormatResponse] = CodeFormatResponse
    _tags = ["formatting"]
    _tool_name = "codeformat"

    def execute(self, request: CodeFormatRequest, metadata: dict) -> CodeFormatResponse:
        results = []
        path = Path(request.path)
        if not path.exists():
            return CodeFormatResponse(
                results=[],
                error=f"The specified path '{request.path}' does not exist or is not accessible.",
            )

        files_to_process = [path] if path.is_file() else list(path.rglob("*.py"))
        for file in files_to_process:
            format_changes = self._run_ruff_format(file)
            errors = self._run_ruff_check(file)
            results.append(
                FormatResult(
                    file_path=str(file),
                    format_changes=format_changes,
                    errors=errors,
                )
            )

        return CodeFormatResponse(results=results)

    def _run_ruff_format(self, file_path: Path) -> List[str]:
        return self._run_ruff_command(file_path, "format")

    def _run_ruff_check(self, file_path: Path) -> List[str]:
        return self._run_ruff_command(file_path, "check")

    def _run_ruff_command(self, file_path: Path, command: str) -> List[str]:
        try:
            result = subprocess.run(
                ["ruff", command, str(file_path)],
                capture_output=True,
                text=True,
                check=True,
            )
            return result.stdout.splitlines()
        except subprocess.CalledProcessError as e:
            return [f"Error running ruff {command}: {e.stderr}"]



from .click_element import ClickElement
from .get_element_details import GetElementDetails
from .get_page_details import GetPageDetails
from .get_screenshot import GetScreenshot
from .goto_page import GotoPage
from .navigate_history import NavigateHistory
from .refresh_page import RefreshPage
from .scroll_page import ScrollPage
from .type_text import TypeText



import typing as t

from pydantic import BaseModel, Field

from composio.tools.local.clickup.actions.base import OpenAPIAction


class GetSpaceViewsRequest(BaseModel):
    """Request schema for `GetSpaceViews`"""

    space_id: int = Field(
        ...,
        alias="space_id",
        description="",
    )


class GetSpaceViewsResponse(BaseModel):
    """Response schema for `GetSpaceViews`"""

    data: t.Dict[str, t.Any]


class GetSpaceViews(OpenAPIAction):
    """View the task and page views available for a Space."""

    _tags = ["Views"]
    _display_name = "get_space_views"
    _request_schema = GetSpaceViewsRequest
    _response_schema = GetSpaceViewsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/space/{space_id}/view"
    method = "get"
    operation_id = "Views_spaceViewsGet"
    action_identifier = "/space/{space_id}/view_get"

    path_params = {"space_id": "space_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}


class views_space_views_get(OpenAPIAction):
    """
    View the task and page views available for a Space.<<DEPRECATED use get_space_views>>
    """

    _tags = ["Views"]
    _display_name = "views_space_views_get"
    _request_schema = GetSpaceViewsRequest
    _response_schema = GetSpaceViewsResponse

    url = "https://api.clickup.com/api/v2"
    path = "/space/{space_id}/view"
    method = "get"
    operation_id = "Views_spaceViewsGet"
    action_identifier = "/space/{space_id}/view_get"

    path_params = {"space_id": "space_id"}
    query_params = {}
    header_params = {}
    request_params = {}

    aliases = {}



"""
Trigger enums.
"""
from composio.client.enums.base import TRIGGERS_CACHE, TriggerData, _AnnotatedEnum, enum

@enum
class Trigger(_AnnotatedEnum[TriggerData], path=TRIGGERS_CACHE):
    """Trigger object."""
    ASANA_TASK_TRIGGER: 'Trigger'
    GITHUB_COMMIT_EVENT: 'Trigger'
    GITHUB_FOLLOWER_EVENT: 'Trigger'
    GITHUB_ISSUE_ADDED_EVENT: 'Trigger'
    GITHUB_LABEL_ADDED_EVENT: 'Trigger'
    GITHUB_PULL_REQUEST_EVENT: 'Trigger'
    GITHUB_STAR_ADDED_EVENT: 'Trigger'
    GMAIL_NEW_GMAIL_MESSAGE: 'Trigger'
    GOOGLEDRIVE_GOOGLE_DRIVE_CHANGES: 'Trigger'
    LINEAR_COMMENT_EVENT_TRIGGER: 'Trigger'
    LINEAR_ISSUE_CREATED_TRIGGER: 'Trigger'
    LINEAR_ISSUE_UPDATED_TRIGGER: 'Trigger'
    MAILCHIMP_MAILCHIMP_CAMPAIGN_TRIGGER: 'Trigger'
    MAILCHIMP_MAILCHIMP_PROFILE_UPDATE_TRIGGER: 'Trigger'
    MAILCHIMP_MAILCHIMP_SUBSCRIBE_TRIGGER: 'Trigger'
    MAILCHIMP_MAILCHIMP_UNSUBSCRIBE_TRIGGER: 'Trigger'
    NOTION_PAGE_ADDED_TO_DATABASE: 'Trigger'
    ONE_DRIVE_ONE_DRIVE_ITEM_TRIGGER: 'Trigger'
    OUTLOOK_OUTLOOK_MESSAGE_TRIGGER: 'Trigger'
    PIPEDRIVE_PIPEDRIVE_NEW_DEAL_TRIGGER: 'Trigger'
    PIPEDRIVE_PIPEDRIVE_NEW_NOTE_TRIGGER: 'Trigger'
    PIPEDRIVE_PIPEDRIVE_NEW_ORGANIZATION_TRIGGER: 'Trigger'
    SLACKBOT_CHANNEL_CREATED: 'Trigger'
    SLACKBOT_REACTION_ADDED: 'Trigger'
    SLACKBOT_REACTION_REMOVED: 'Trigger'
    SLACKBOT_RECEIVE_MESSAGE: 'Trigger'
    SLACKBOT_RECEIVE_THREAD_REPLY: 'Trigger'
    SLACK_CHANNEL_CREATED: 'Trigger'
    SLACK_REACTION_ADDED: 'Trigger'
    SLACK_REACTION_REMOVED: 'Trigger'
    SLACK_RECEIVE_MESSAGE: 'Trigger'
    SLACK_RECEIVE_THREAD_REPLY: 'Trigger'
    TRELLO_TRELLO_ARCHIVED_CARD_TRIGGER: 'Trigger'
    TRELLO_TRELLO_NEW_BOARD_TRIGGER: 'Trigger'
    TRELLO_TRELLO_NEW_CARD_TRIGGER: 'Trigger'
    TRELLO_TRELLO_UPDATED_CARD_TRIGGER: 'Trigger'
    YOUTUBE_NEW_ACTIVITY_TRIGGER: 'Trigger'
    YOUTUBE_NEW_ITEM_IN_PLAYLIST_TRIGGER: 'Trigger'
    YOUTUBE_NEW_PLAYLIST_TRIGGER: 'Trigger'
    YOUTUBE_NEW_SUBSCRIPTION_TRIGGER: 'Trigger'
    ZENDESK_NEW_USER_TRIGGER: 'Trigger'
    ZENDESK_NEW_ZENDESK_TICKET_TRIGGER: 'Trigger'

    @property
    def name(self) -> str:
        """Name of the trigger."""
        return self.load().name

    @property
    def app(self) -> str:
        """Name of the app where this trigger belongs to."""
        return self.load().app


import traceback

import sentry_sdk


def CatchAllExceptions(cls, handler):
    class Cls(cls):
        _original_args = None

        # pylint: disable=unused-argument
        def make_context(self, info_name, args, parent=None, **extra):
            # grab the original command line arguments
            self._original_args = " ".join(args)

            try:
                # pylint: disable=super-with-arguments
                return super(Cls, self).make_context(
                    info_name, args, parent=parent, **extra
                )
                # pylint: enable=super-with-arguments
            except Exception as exc:
                # call the handler
                should_ignore_error = handler(self, info_name, exc)
                if should_ignore_error:
                    return None
                # let the user see the original error
                raise

        # pylint: enable=unused-argument

        # pylint: disable=unused-argument
        def invoke(self, ctx):
            try:
                # pylint: disable=super-with-arguments
                return super(Cls, self).invoke(ctx)
                # pylint: enable=super-with-arguments
            except Exception as exc:
                # call the handler
                should_ignore_error = handler(self, ctx.info_name, exc)
                if should_ignore_error:
                    return None
                # let the user see the original error
                raise

        # pylint: enable=unused-argument

    return Cls


# pylint: disable=unused-argument
def handle_exceptions(cmd, info_name, exc):
    # send error info to rollbar, etc, here
    if (isinstance(exc, (SystemExit, ValueError))) and sentry_sdk.is_initialized():
        sentry_sdk.capture_exception(exc)
        sentry_sdk.flush()
        print(traceback.format_exc())
        return True

    return False


# pylint: enable=unused-argument
